# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Usage
=====

(as inferred from the source by Michael Droettboom)

See the docstrings of the classes and functions for more information.

.. code::

  from gamera import roman_text

  # Load an image and a classifier (see documentation elsewhere)...

  # image is a one-bit image
  # classifier is a classifier object with a production database loaded
  
  page = roman_text.ocr(image, classifier)

*page* is a ``Page`` object.  You can display how the page was
segmenteted in different ways using ``.display_sections``,
``.display_lines``, ``.display_glyphs``, and
``.display_segmentation``.

Once you have a segmented page object, you can convert it to ASCII
using the ``make_string`` function.  It takes as input a ``Section``
object, and an optional name-mapping function.  The name-mapping
function is any function that takes a string with a symbol id_name and
returns a character (which may be Unicode).  This function defines how
the names used in the training data map to characters in the output.

If no name-mapping function is provided a default one is used which
maps Gamera symbol id names to Unicode character names and then to
Unicode characters.  All periods ('.') in the Gamera id name are
converted to spaces, and then the result is used as a Unicode
character name.  For instance ``latin.capital.letter.a`` maps to
``LATIN CAPITAL LETTER A``.  (`Charts of Unicode character names`_).
The version of Unicode used corresponds with that used in your version
of Python.  For instance, Python 2.3.x supports Unicode 3.2.0.

.. _: http://www.unicode.org/charts/

.. code::

  output = []
  for section in page.sections:
    output.append(roman_text.make_string(section))

"""

from gamera import core
import unicodedata
import string

class Page:
    def __init__(self, image, glyphs):
        """Create a page with the given image."""
        self.sections = []
        self.image = image
        self.glyphs = glyphs
        self.section_search_size = 1
        self.__fill = 0
        self.__noise_size = 0
        self.__large_size = 0

    def __section_size_test(self, glyph):
        """Filter for section finding - removes very small and
        very large glyphs"""
        black_area = glyph.black_area()[0]
        if black_area > self.__noise_size and \
                glyph.nrows < self.__large_size and \
                glyph.ncols < self.__large_size:
            return 1
        else:
            return 0

    def __avg_glyph_size(self, glyphs):
        """Compute the average glyph size for the page"""
        total = 0.0
        for g in glyphs:
            total += g.nrows
            total += g.ncols
        return total / (2 * len(glyphs))

    def __find_intersecting_rects(self, glyphs, index):
        """For section finding - return the index of glyphs intersecting
        the glyph and the index passed in."""
        g = glyphs[index]
        inter = []
        for i in range(len(glyphs)):
            if i == index:
                continue
            if g.intersects(glyphs[i]):
                inter.append(i)
        return inter

    def segment(self):
        """Segment the page into sections and lines. Also computes
        all of the statistics for the sections and lines."""
        self.find_sections()
        self.find_lines()
        for s in self.sections:
            s.calculate_stats()
            for line in s.lines:
                line.calculate_stats()

    def find_lines(self):
        """Find the lines within all of the sections. Must be called
        after find_sections."""
        for s in self.sections:
            s.find_lines()

    def find_sections(self):
        """Find the sections within an image - this finds large blocks
        of text making it possible to find the lines within complex
        text layouts."""
      
        glyphs = self.glyphs

        FUDGE = self.__avg_glyph_size(glyphs) * self.section_search_size

        # remove noise and large objects
        self.__noise_size = FUDGE
        self.__large_size = FUDGE * 20
        new_glyphs = []
        for g in glyphs:
            if self.__section_size_test(g):
                new_glyphs.append(g)
            else:
                if self.__fill:
                    g.fill_white()
        glyphs = new_glyphs
        
        # Sort the glyphs left-to-right and top-to-bottom
        glyphs.sort(lambda x, y: cmp(x.ul_x, y.ul_x))
        glyphs.sort(lambda x, y: cmp(x.ul_y, y.ul_y))
        
        # Create rectangles for each glyph that are bigger by FUDGE
        big_rects = []
        for g in glyphs:
            ul_y = max(0, g.ul_y - FUDGE)
            ul_x = max(0, g.ul_x - FUDGE)
            lr_y = min(self.image.lr_y, g.lr_y + FUDGE)
            lr_x = min(self.image.lr_x, g.lr_x + FUDGE)
            ul_x = int(ul_x); ul_y = int(ul_y)
            nrows = int(lr_y - ul_y + 1)
            ncols = int(lr_x - ul_x + 1)
            big_rects.append(core.Rect(core.Point(ul_x, ul_y), core.Dim(ncols, nrows)))

        # Search for intersecting glyphs and merge them. This is
        # harder than it seems at first because we want everything
        # to merge together that intersects regardless of the order
        # in the list. It ends up being similar to connected-component
        # labeling. This is prone to be kind-of slow.
        current = 0
        rects = big_rects
        while(1):
            # Find the indexexes of any rects that interesect with current
            inter = self.__find_intersecting_rects(rects, current)
            # If we found intersecting rectangles merge them with them current
            # rect, remove them from the list, and start the whole process
            # over. We start over to make certain that everything that should
            # be merged is.
            if len(inter):
                g = rects[current]
                new_rects = [g]
                for i in range(len(rects)):
                    if i == current:
                        continue
                    if i in inter:
                        g.union(rects[i])
                    else:
                        new_rects.append(rects[i])
                rects = new_rects
                current = 0
            # If we didn't find anything that intersected move on to the next
            # rectangle.
            else:
                current += 1
            # Bail when we are done.
            if current >= len(rects):
                break
        
        # Create the sections
        sections = []
        for rect in rects:
            sections.append(Section(rect))
        
        # Place the original (small) glyphs into the sections
        for glyph in self.glyphs:
            if self.__section_size_test(glyph):
                for s in sections:
                    if s.bbox.intersects(glyph):
                        s.add_glyph(glyph)
                        break
        
        # Fix up the bounding boxes
        for s in sections:
            s.calculate_bbox()
                
        self.sections = sections
        
    def display_sections(self, clear=1):
        """Display the sections found by placing a box around them
        in a display. If clear is true then any boxes already on
        the displayed are cleared first."""
        # display the sections
        result = self.sections                

        if self.image._display == None:
            self.image.display()
        display = self.image._display
        if clear:
            display.clear_all_boxes()
        for rect in result:
            b = rect.bbox
            display.add_box(core.Rect(core.Point(b.ul_x, b.ul_y), core.Dim(b.ncols, b.nrows)))

    def display_lines(self, clear=1):
        """Display the lines found by placing a box around them
        in a display. If clear is true then any boxes already on
        the displayed are cleared first."""
        if self.image._display == None:
            self.image.display()
        display = self.image._display
        if clear:
            display.clear_all_boxes()
        for s in self.sections:
            for line in s.lines:
                b = line.bbox
                display.add_box(core.Rect(core.Point(b.ul_x, b.ul_y), core.Dim(b.ncols, b.nrows)))

    def display_glyphs(self, clear=1):
        if self.image._display == None:
            self.image.display()
        display = self.image._display
        if clear:
            display.clear_all_boxes()
        for rect in self.glyphs:
            b = rect
            display.add_box(b.ul_y, b.ul_x, b.nrows, b.ncols)

    def display_segmentation(self):
        self.display_lines()
        self.display_sections(0)

class Section:
    def __init__(self, bbox):
        self.bbox = core.Rect(bbox)
        self.lines = []
        self.glyphs = []
        # stats
        self.avg_glyph_area = 0
        self.avg_glyph_height = 0
        self.avg_glyph_width = 0
        self.avg_line_height = 0
        self.agv_line_width = 0

    def find_tall_glyphs(self, stdev=20):
        from gamera import stats
        tall = []
        for i in range(len(self.glyphs)):
            g = self.glyphs[i]
            if stats.samplestdev([g.nrows, self.avg_glyph_height]) > stdev:
                tall.append(i)
        return tall

    def add_line(self, line):
        self.lines.append(line)
        self.lines.sort(lambda x, y: cmp(x.bbox.ul_y, y.bbox.ul_y))

    def add_glyph(self, glyph):
        self.glyphs.append(glyph)

    def calculate_bbox(self):
        assert(len(self.glyphs) > 0)
        self.bbox = core.Rect(self.glyphs[0])
        for glyph in self.glyphs:
            self.bbox.union(glyph)

    def calculate_stats(self):
        self.calculate_glyph_stats()
        self.calculate_line_stats()

    def calculate_glyph_stats(self):
        # calculate glyph stats
        total_area = 0.0
        total_gheight = 0.0
        total_gwidth = 0.0
        
        for g in self.glyphs:
            nrows = g.nrows
            ncols = g.ncols
            total_area += nrows * ncols
            total_gheight += nrows
            total_gwidth += ncols
        l = len(self.glyphs)
        self.avg_glyph_area = total_area / l
        self.avg_glyph_height = total_gheight / l
        self.avg_glyph_width = total_gwidth / l

    def calculate_line_stats(self):
        # calculate line stats
        total_lheight = 0.0
        total_lwidth = 0.0
        for line in self.lines:
            total_lheight += line.bbox.nrows
            total_lwidth += line.bbox.ncols
        l = len(self.lines)
        self.avg_line_height = total_lheight / l
        self.avg_line_width = total_lwidth / l
        

    def __find_intersecting_lines(self, glyphs, index):
        g = glyphs[index]
        inter = []
        for i in range(len(glyphs)):
            if i == index:
                continue
            if g.bbox.intersects(glyphs[i].bbox):
                inter.append(i)
        return inter

    def find_lines(self):
        # Remove abnormally tall glyphs that might interfer with
        # line finding
        self.calculate_glyph_stats()
        tall_indexes = self.find_tall_glyphs()
        tall = []
        glyphs = []
        for i in range(len(self.glyphs)):
            if i in tall_indexes:
                tall.append(self.glyphs[i])
            else:
                glyphs.append(self.glyphs[i])
        orig_glyphs = self.glyphs
        self.glyphs = glyphs
        
        # find the lines - this is very basic for now
        lines = []
        for glyph in self.glyphs:
            found = 0
            for line in lines:
                if line.contains_glyph(glyph):
                    line.add_glyph(glyph)
                    found = 1
                    break
            if not found:
                new_line = Line(glyph)
                lines.append(new_line)

        # Merge any overlapping lines
        current = 0
        while(1):
            inter = self.__find_intersecting_lines(lines, current)
            if len(inter):
                new_lines = [lines[current]]
                for i in range(len(lines)):
                    if i == current:
                        continue
                    if i in inter:
                        new_lines[0].merge(lines[i])
                    else:
                        new_lines.append(lines[i])
                current = 0
                lines = new_lines
            else:
                current += 1
            if current >= len(lines):
                break

        # Put the tall glyphs back in by assigning them to the first line
        # we come to that intersects.
        for i in range(len(tall)):
            for line in lines:
                found = 0
                if line.bbox.contains_y(tall[i].ul_y):
                    line.add_glyph(tall[i])
                    found = 1
                    break
            if not found:
                print "Did not find lines for all tall glyphs"
                
        self.glyphs = orig_glyphs
        self.lines = lines
        

class Line:
    def __init__(self, glyph):
        self.center = 0
        self.bbox = core.Rect(core.Point(glyph.ul_x, glyph.ul_y), core.Dim(glyph.ncols, glyph.nrows))
        self.glyphs = []
        self.add_glyph(glyph)

    def add_glyph(self, glyph):
        self.glyphs.append(glyph)
        self.glyphs.sort(lambda x, y: cmp(x.ul_x, y.ul_x))
        self.bbox.union(glyph)

    def calculate_stats(self):
        total_center = 0
        for glyph in self.glyphs:
            total_center += (glyph.ul_y + glyph.lr_y) / 2
            self.bbox.union(glyph)
        self.center = total_center / len(self.glyphs)

    def contains_glyph(self, glyph):
        center = (glyph.ul_y + glyph.lr_y) / 2
        if self.bbox.contains_y(center):
            return 1
        else:
            return 0

    def merge(self, line):
        self.glyphs.extend(line.glyphs)
        self.calculate_stats()
        
def name_lookup_old(id_name):
    """Converts a symbol name into a single character."""
    split_string = string.split(id_name, '.')
    l = len(split_string)
    if l > 1:
        if split_string[0] == "lower":
            return string.lower(split_string[-1])
        elif split_string[0] == "upper":
            return string.upper(split_string[-1])
        elif split_string[0] == "symbol":
            if l == 2:
                s = split_string[1]
                if s == "comma":
                    return ","
                elif s == "dot":
                    return ""
                elif s == "plus":
                    return "+"
                elif s == "lessthan":
                    return "<"
                else:
                    print "ERROR: Name not known about:", id_name
                    return ""
            elif l == 3:
                if split_string[1] == "paren":
                    if split_string[2] == "close":
                        return ")"
                    else:
                        return "("
                elif split_string[1] == "bracket":
                    if split_string[2] == "close":
                        return "]"
                    else:
                        return "["
                else:
                    print "ERROR: Name not known about:", id_name
                    return ""
            else:
                print "ERROR: Name not known about:", id_name
                return ""
        elif split_string[0] == "digit":
            return split_string[1]
        elif split_string[0] == "line":
            return ""
        elif split_string[0] == "punctuation":
            if split_string[1] == "comma":
                return ","
            else:
                return ""
        else:
            print "ERROR: Name not known about:", id_name
            return ""
    else:
        return ""

def name_lookup_unicode(id_name):
    name = id_name.replace(".", " ")
    name = name.upper()
    try:
        return unicodedata.lookup(name)
    except KeyError:
        print "ERROR: Name not found:", name
        return ""
    
def make_string(lines, name_lookup_func=name_lookup_unicode):
    s = ""

    for line in lines:
        glyphs = line.glyphs
        total_space = 0
        for i in range(len(glyphs) - 1):
            total_space += glyphs[i + 1].ul_x - glyphs[i].lr_x
        average_space = total_space / len(glyphs)
        for i in range(len(glyphs)):
            if i > 0:
                if (glyphs[i].ul_x - glyphs[i - 1].lr_x) > (average_space * 2):
                    
                    s = s + " "
            s = s + name_lookup_func(glyphs[i].get_main_id())
        s = s + "\n"
    return s

def output(lines, filename="text.txt"):
    s = make_string(lines)
    f = file(filename, "w")
    f.write(s)
    f.flush()
    f.close()

def ocr(image, classifier = None, glyphs = None):
    """Performs basic OCR segmentation.

*image*: OneBit image
*classifier*: a classifier with a production database loaded (optional)
*glyphs*: a list of glyphs from the image (optional)

Returns a ``Page`` object.
"""
    if glyphs is None:
        glyphs = image.cc_analysis()
    if classifier is not None:
        #classifier.classify_list_automatic(glyphs)
        glyphs = classifier.group_and_update_list_automatic(glyphs)
    page = Page(image, glyphs)
    page.segment()
    return page



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2009 Ichiro Fujinaga, Michael Droettboom,
#                         Karl MacMillan, and Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.args import *
from gamera import paths
from gamera import util
import new, os, os.path, imp, inspect, sys, copy
from gamera.backport import sets
from types import *
from enums import *


plugin_methods = {}

class PluginModule:
   category = None
   cpp_namespaces = []
   cpp_sources = []
   cpp_headers = []
   cpp_defines = []
   cpp_include_dirs = []
   extra_libraries = []
   library_dirs = []
   define_macros = []
   extra_compile_args = []
   extra_link_args = []
   extra_objects = []
   functions = []
   pure_python = False
   version = "1.0"
   author = None
   url = None
   stable = True

   def __init__(self):
      for function in self.functions:
         function.module = self.__class__
      core = __import__("gamera.core")
      if core is None:
         return
      for function in self.functions:
         if not isinstance(function, ClassType):
            function = function.__class__
         function.register()

class Builtin(PluginModule):
   author = "Michael Droettboom and Karl MacMillan"
   url = "http://dkc.jhu.edu/"

class PluginFunction:
   return_type = None
   self_type = ImageType(ALL)
   args = Args([])
   image_types_must_match = 0
   testable = 0
   feature_function = False
   doc_examples = []
   category = None
   pure_python = False
   progress_bar = ""
   author = None
   add_to_image = True

   def get_formatted_argument_list(cls):
      return "**%s** (%s)" % (cls.__name__, ', '.join(
         [x.rest_repr(True) for x in cls.args.list]))
   get_formatted_argument_list = classmethod(get_formatted_argument_list)

   def escape_docstring(cls):
      if cls.__doc__ is None:
         doc = ''
      doc = util.dedent(cls.__doc__)
      doc = doc.replace("\n", r"\n").replace('"', r'\"')
      return r'"%s\n\n%s"' % (cls.get_formatted_argument_list(), doc)
   escape_docstring = classmethod(escape_docstring)

   def register(cls):
      # add_to_image = add_to_image and cls.add_to_image
      if cls.return_type != None:
         if cls.return_type.name == None:
            cls.return_type = copy.copy(cls.return_type)
            cls.return_type.name = cls.__name__
      if not hasattr(cls, "__call__"):
         # This loads the actual C++ function if it is not directly
         # linked in the Python PluginFunction class
         parts = cls.__module__.split('.')
         file = inspect.getfile(cls)
         cpp_module_name = '_' + parts[-1]
         directory = os.path.split(file)[0]
         sys.path.append(directory)
         found = imp.find_module(cpp_module_name)
         del sys.path[-1]
         if found:
             module = imp.load_module(cpp_module_name, *found)
         if module == None:
            return
         func = getattr(module, cls.__name__)
      elif cls.__call__ is None:
         func = None
      else:
         func = cls.__call__
         if type(cls.__call__) == new.instancemethod:
            func = cls.__call__.im_func
         func.func_doc = ("%s\n\n%s" %
                          (cls.get_formatted_argument_list(),
                           util.dedent(cls.__doc__)))
      cls.__call__ = staticmethod(func)

      if cls.category == None:
         category = cls.module.category
      else:
         category = cls.category
      if not category is None and not func is None:
         image_type = isinstance(cls.self_type, ImageType)
         if image_type:
            pixel_types = cls.self_type.pixel_types
         else:
            pixel_types = [NONIMAGE]
         for pixel_type in pixel_types:
            if not plugin_methods.has_key(pixel_type):
               plugin_methods[pixel_type] = {}
            start = plugin_methods[pixel_type]
            for subcategory in category.split('/'):
               if not start.has_key(subcategory):
                  start[subcategory] = {}
               start = start[subcategory]
            start[cls.__name__] = cls

      if not func is None and not cls.self_type is None:
         cls.self_type.register(cls, func)
   register = classmethod(register)

def PluginFactory(name, category=None,
                  return_type=None,
                  self_type=ImageType(ALL),
                  args=None):
   from gamera import core
   cls = new.classobj(name, (PluginFunction,), {})
   if not category is None:
      cls.category = category
   cls.return_type = return_type
   cls.self_type = self_type
   if args is None:
      cls.args = Args([])
   else:
      cls.args = args
   func = getattr(core.ImageBase, name)
   cls.__call__ = func
   cls.__doc__ = util.dedent(func.__doc__)
   cls.module = Builtin
   return cls

def get_config_options(command):
   return os.popen(command).read()

def methods_flat_category(category, pixel_type=None):
   if pixel_type == None:
      methods = sets.Set()
      for pixel_type in ALL + [NONIMAGE]:
         # We have to cast the lists to sets here to make Python 2.3.0 happy.
         methods.update(sets.Set(methods_flat_category(category, pixel_type)))
      return list(methods)
   elif plugin_methods.has_key(pixel_type):
      methods = plugin_methods[pixel_type]
      if methods.has_key(category):
         return _methods_flatten(methods[category])
   return []

def _methods_flatten(mat):
   list = []
   for key, val in mat.items():
      if type(val) == DictType:
         list.extend(_methods_flatten(val))
      else:
         list.append((key, val))
   return list



#
# various editing techniques to improve kNN Classifier performance
# Copyright (C) 2007, 2008 Colin Baumgarten
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from random import randint
from gamera.args import Args, Int, Check
from gamera.knn import kNNInteractive
from gamera.util import ProgressFactory

def _randomSetElement(set):
    """Retrieve a random element from the given set

    *set*
      The set from which to select a random element"""
    randPos = randint(0, len(set) - 1);
    for pos, elem in enumerate(set):
        if pos == randPos:
            return elem

def _copyClassifier(original, k = 0):
    """Copy a given kNNClassifer by constructing a new one with identical
parameters.

    *original*
      The classifier to be copied

    *k*
      If the copy shall have another k-value as the original, set k accordingly.
      k = 0 means, that the original's k-value will be used"""
    if k == 0:
        k = original.num_k
    return kNNInteractive(
                list(original.get_glyphs()), original.features,
                original._perform_splits,
                k)

def _getMainId(classificationResult):
    """Classification results returned from a kNN Classifier are in a list
containing '(confidence, className)' tuples. So to determine the 'main class',
the classname with the highest confidence has to be returned.
"""
    # classificationResult is a tuple of the form
    #   ([idname1, idname2, ...], confidences_for_idname1)
    return classificationResult[0][0][1]

class AlgoRegistry(object):
    """Registry containing a list of all available editing algorithms. Besides
the callable itself, the registry also stores its docstring, a displayname and
type-information about any additionally required arguments. This information
will be used by the gui to show a dialog to execute any of the available
algorithms on a classifier.

An editing algorithm is any callable object, that takes at least one parameter
-  a *gamera.knn.kNNInteractive* classifier - and returns a new edited
*kNNInteractive* classifier.

To add your own algorithm, use one of the *register()* methods or alternatively,
let your callable inherit from *EditingAlgorithm* to register your algorithm,"""
    algorithms = []

    def registerData(algoData):
        """Register a new editing Algorithm using metadata from an
*AlgoData* object"""
        AlgoRegistry.algorithms.append(algoData)
    registerData = staticmethod(registerData)

    def register(name, callable, args = Args(), doc = ""):
        """Register a new editing Algorithm: The parameters are the same as in
*AlgoData.__init__*, so see its doc for an explanation of the parameters."""
        AlgoRegistry.registerData(AlgoData(name, callable, args, doc))
    register = staticmethod(register)

class AlgoData(object):
    """Class holding all metadata about an editing algorithm, that is required by
*AlgoRegistry*

*name*
    Name of the algorithm

*callable*
    The callable object implementing the algorithm. Its first parameter has to
    be a *kNNInteractive* Classifier and it has to return a new *kNNInteractive*
    classifier. If the algorithm requires any additional parameters, they have
    to be specified in the *args* parameter

*args*
    A *gamera.args.Args* object specifying any additional parameters required by
    the algorithm

*doc*
    A docstring in reStructured Text format, describing the algorithm and its
    parameters"""
    def __init__(self, name, callable, args = Args(), doc = ""):
        self.name = name
        self.callable = callable
        self.args = args
        self.doc = doc


class EditingAlgorithm(object):
    """Convenience class to automatically register editing algorithms with the
*AlgoRegistry*. If you implement your algorithm as a callable class, you can
just inherit from this class to let it automatically be registered. Just add
two properties to the class:

*name*
    The name of your algorithm
*args*
    Type info about any additional required arguments (a *gamera.args.Args* object)
"""
    def __init__(self):
        AlgoRegistry.register(self.name, self, self.args, self.__doc__)

class EditMnn(EditingAlgorithm):
    """**edit_mnn** (kNNInteractive *classifier*, int *k* = 0, bool *protectRare*, int *rareThreshold*)
    
Wilson's *Modified Nearest Neighbour* (MNN, aka *Leave-one-out-editing*).
The algorithm removes 'bad' glyphs from the classifier, i.e. glyphs that
are outliers from their class in featurespace, usually because they have been 
manually misclassified or are not representative for their class

    *classifier*
        The classifier from which to create an edited copy
    *internalK*
        The k value used internally by the editing algorithm. If 0 is given for 
        this parameter, the original classifier's k is used (recommended).        
    *protect rare classes*
        The algorithm tends to completely delete the items of rare classes,
        removing this whole class from the classifier. If this is not 
        desired these rare classes can be explicitly protected from deletion.
        Note that enabling this option causes additional computing effort    
    *rare class threshold*
        In case *protect rare classes* is enabled, classes with less than this
        number of elements are considered to be rare

Reference: D. Wilson: 'Asymptotic Properties of NN Rules Using Edited Data'.
*IEEE Transactions on Systems, Man, and Cybernetics*, 2(3):408-421, 1972
"""
    name = "Wilson's Modified Nearest Neighbour (MNN)"
    args = Args([Int("Internal k", default = 0),
                 Check("Protect rare classes", default = True),
                 Int("Rare class threshold", default = 3)])

    def __call__(self, classifier, k = 0, protectRare = True,
                 rareThreshold = 3):

        editedClassifier = _copyClassifier(classifier, k)
        toBeRemoved = set()
        progress = ProgressFactory("Generating edited MNN classifier...",
                                      len(classifier.get_glyphs()))

        # classify each glyph with its leave-one-out classifier
        for i, glyph in enumerate(classifier.get_glyphs()):
            editedClassifier.get_glyphs().remove(glyph)
            detectedClass = _getMainId(\
                                editedClassifier.guess_glyph_automatic(glyph))
            # check if recognized class complies with the true class
            if glyph.get_main_id() != detectedClass:
                toBeRemoved.add(glyph)
            editedClassifier.get_glyphs().add(glyph)
            progress.step()

        rareClasses = self._getRareClasses(classifier.get_glyphs(),
                                           protectRare, rareThreshold)
        
        # remove 'bad' glyphs, if they are not in a rare class
        for glyph in toBeRemoved:
            if glyph.get_main_id() in rareClasses:
                continue
            editedClassifier.get_glyphs().remove(glyph)
            
        progress.kill()
        return editedClassifier
    
    def _getRareClasses(self, glyphs, protectRare, rareThreshold):
        """Produces a set containing the names of all rare classes"""
        result = set()
        
        if not protectRare: 
            return result
        
        # histogram of classNames
        histo = {}        
        for g in glyphs:
            count = histo.get(g.get_main_id(), 0)
            histo[g.get_main_id()] = count + 1
        
            
        for className in histo:
            if histo[className] < rareThreshold:
                result.add(className)        
        return result

edit_mnn = EditMnn()    

class EditCnn(EditingAlgorithm):
    """**edit_cnn** (kNNInteractive *classifier*, int *k* = 0, bool *randomize*)
    
Hart's *Condensed Nearest Neighbour (CNN)* editing.
This alorithm is specialized in removing superfluous glyphs - glyphs that do not
influence the recognition rate - from the classifier to improve its 
classification speed. Typically glyphs far from the classifiers decision 
boundaries are removed.

    *classifier*
        The classifier from which to create an edited copy
    *internalK*
        The k value used internally by the editing algorithm. 0 means, use the 
        same value as the given classifier (recommended)    
    *randomize*
        Because the processing order of the glyphs in the classifier impacts the
        result of this algorithm, the order will be randomized. If reproducible
        results are required, turn this option off.

Reference: P.E. Hart: 'The Condensed Nearest Neighbor rule'. *IEEE Transactions on Information Theory*, 14(3):515-516, 1968
"""
    name = "Hart's Condensed Nearest Neighbour (CNN)"
    args = Args([Int("Internal k", default=0), 
                 Check("Randomize", default = True)])

    def __call__(self, classifier, k = 0, randomize = True):
        # special case of empty classifier
        if (not classifier.get_glyphs()):
            return _copyClassifier(classifier)

        if k == 0:
            k = classifier.num_k
        
        progress = ProgressFactory("Generating edited CNN classifier...",
                                      len(classifier.get_glyphs()))

        # initialize Store (a) with a single element
        if randomize:
            elem = _randomSetElement(classifier.get_glyphs())
        else:
            elem = classifier.get_glyphs().__iter__().next()
        
        aGlyphs = [elem]
        a = kNNInteractive(aGlyphs, classifier.features, 
                           classifier._perform_splits, k)
        progress.step()
        
        # initialize Grabbag (b) with all other
        b = classifier.get_glyphs().copy()
        b.remove(aGlyphs[0]);

        # Classify each glyph in b with a as the classifier
        # If glyph is misclassified, add it to a, repeat until no elements are 
        # added to a
        changed = True
        while changed == True:
            changed = False
            # copy needed because iteration through dict is not possible while 
            # deleting items from it
            copyOfB = b.copy()
            for glyph in copyOfB:
                if glyph.get_main_id() != _getMainId(a.guess_glyph_automatic(glyph)):
                    b.remove(glyph)
                    a.get_glyphs().add(glyph)
                    progress.step()
                    changed = True
        progress.kill()
        a.num_k = 1
        return a

edit_cnn = EditCnn()

class EditMnnCnn(EditingAlgorithm):
    """**edit_mnn_cnn** (kNNInteractive *classifier*, int *k* = 0, bool *protectRare*, int *rareThreshold*, bool *randomize*)

Combined execution of Wilson's Modified Nearest Neighbour and Hart's
Condensed Nearest Neighbour. Combining the algorithms in this order is 
recommended, because first bad samples are removed to improve the classifiers 
accuracy, and then the remaining samples are condensed to speed up the classifier

For documentation of the parameters see the independent algorithms"""
    name = "MNN, then CNN"
    args = Args([Int("Internal k", default = 0),
                 Check("Protect rare classes", default = True),
                 Int("Rare class threshold", default = 3),
                 Check("Randomize", default = True)])

    def __call__(self, classifier, k = 0, protectRare = True,
                 rareThreshold = 3, randomize = True):
        return edit_cnn(edit_mnn(classifier, k, protectRare, rareThreshold),
                   randomize)

edit_mnn_cnn = EditMnnCnn()



import sys

def my_application():
   image = load_image(sys.argv[-1])
   onebit = image.otsu_threshold()

   # Display the image in a window
   onebit.display()	       

   # Assign it do a global variable so the window doesn't
   # disappear
   global onebit_global
   onebit_global = onebit

# Import the Gamera core and initialise it
from gamera.core import *
init_gamera()

# Import the Gamera GUI and start it   
from gamera.gui import gui
gui.run(my_application)



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2014      Fabian Schmitt
#               2009-2014 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _listutilities

class permute_list(PluginFunction):
   """
   Permutes the given list (in place) one step.

   Returns ``True`` if there are more permutations to go.  Returns
   ``False`` if permutations are done.

   Example usage::

     >>>from gamera.plugins import listutilities
     >>>a = [1,2,3]
     >>>while listutilities.permute_list(a):
     ...    print a
     ...
     [2, 1, 3]
     [1, 3, 2]
     [3, 1, 2]
     [2, 3, 1]
     [3, 2, 1]
   """
   category = "List"
   self_type = None
   args = Args([Class("list")])
   return_type = Int("continuaton")

class all_subsets(PluginFunction):
   """Returns all subsets of size *size* of the given list."""
   category = "List"
   self_type = None
   args = Args([Class("list"), Int("size")])
   return_type = Class("subsets")

class median(PluginFunction):
    """Compute the median from a list of values in linear time.

This implementation works both with builtin numeric types like *int* or
*float*, and with user defined types. For user defined type, you
must implement the "less than" operator (`__lt__`), as in the following
example:

.. code:: Python

   class P:
      x = 0; y = 0
      def __init__(self, x, y):
          self.x = x; self.y = y
      def __lt__(self, other):
          return (self.x < other.x)
      def __eq__(self, other):
          return (self.x == other.x)

   a = [P(0,0), P(1,1), P(2,0)]
   p = median(a)

When the parameter *inlist* is ``True``, the median is always a list entry,
even for lists of even size. Otherwise, the median is for an
even size list of *int* or *float* values the mean between the two middle
values. So if you need the median for a pivot element, set *inlist* to
``True``.

For user defined types, the returned median is always a list
entry, because arithmetic computations do not make sense in this case.

.. note::

   This is *not* the median image filter that replaces each pixel value
   with the median of its neighborhood. For this purpose, see the
   rank__ plugin.

.. __: morphology.html#rank
"""
    category = "List"
    pure_python = 1
    self_type = None
    return_type = Class("m")
    args = Args([Class("list"),
                 Check("inlist", check_box="always from list", default=False)])
    author = "Christoph Dalitz"
    def __call__(list, inlist=False):
        return _listutilities.median_py(list, inlist)
    __call__ = staticmethod(__call__)

class median_py(PluginFunction):
    """This is only for Gamera's Python-C++ interface."""
    category = None
    self_type = None
    return_type = Class("m")
    args = Args([Class("list"), Check("inlist")])
    author = "Christoph Dalitz"

class kernel_density(PluginFunction):
    """Computes the kernel density for *values* at the specified
*x*-positions. Reference: S.J. Sheather: \"Density Estimation.\"
Statistical Science 19 pp. 588-597 (2004).

Arguments:

  *values*
     Sample values from which the density is to be estimated.

  *x*
     For each value in *x*, the density at this position is returned
     in the returned float vector.

  *bw*
     Band width, i.e. the parameter *h* in the kernel density estimator.
     when set to zero, Silverman's rule of thumb is used, which sets the
     bandwidth to 0.9 min{sigma, iqr/1.34} n^(-1/5).

  *kernel*
     The kernel function that weights the values (0 = rectangular, 
     1 = triangular, 2 = Gaussian). A Gaussian kernel produces the smoothest
     result, but is slightly slower than the other two.

     Note that the kernels are normalized to variance one, which means that
     the rectangular kernel has support [-sqrt(3), +sqrt(3)], and the
     triangular kernel has support [-sqrt(6), sqrt(6)].
"""
    category = "List"
    self_type = None
    return_type = FloatVector()
    args = Args([FloatVector("values"), FloatVector("x"), Float("bw", default=0.0), Choice("kernel", choices=["rectangular", "triangular", "gaussian"], default=0)])
    author = "Christoph Dalitz and Fabian Schmitt"
    def __call__(values, x, bw=0.0, kernel=0):
        return _listutilities.kernel_density(values, x, bw, kernel)
    __call__ = staticmethod(__call__)


class argmax(PluginFunction):
    """Returns the index of the maximum in a list.
"""
    category = "List"
    pure_python = 1
    self_type = None
    return_type = Int()
    args = Args([FloatVector("x")])
    author = "Christoph Dalitz"
    def __call__(x):
        mi = 0
        mv = x[0]
        for i,v in enumerate(x):
            if v > mv:
                mi = i; mv = v
        return mi
    __call__ = staticmethod(__call__)

class argmin(PluginFunction):
    """Returns the index of the minimum in a list.
"""
    category = "List"
    pure_python = 1
    self_type = None
    return_type = Int()
    args = Args([FloatVector("x")])
    author = "Christoph Dalitz"
    def __call__(x):
        mi = 0
        mv = x[0]
        for i,v in enumerate(x):
            if v < mv:
                mi = i; mv = v
        return mi
    __call__ = staticmethod(__call__)

class ListUtilitiesModule(PluginModule):
   category = None
   cpp_headers=["listutilities.hpp"]
   functions = [permute_list, all_subsets, median, median_py,
                kernel_density, argmax, argmin]
   author = "Michael Droettboom and Karl MacMillan"
   url = "http://gamera.sourceforge.net/"
module = ListUtilitiesModule()

permute_list = permute_list()
all_subsets = all_subsets()
median = median()
kernel_density = kernel_density()
argmax = argmax()
argmin = argmin()



import py.test

from gamera.core import *
init_gamera()

import features

eps = 0.0001

# testing feature_plugins

# area of bounding box, nrows * ncols, result floatvector[value]
# bounding box of whole pic, independent from content, trivial case
def test_area():
    
    #case 1
    img = Image((0,0), (8,8), ONEBIT)
    area_f = img.area()
    assert area_f[0] == 81.0
    
    #case 2
    img = Image((0,0), (3,5), ONEBIT)
    area_f = img.area()
    assert area_f[0] == 24.0
    
    #case 3
    img = Image((0,0), (3,16), ONEBIT)
    area_f = img.area()
    assert area_f[0] == 68.0
    
    
# aspect_ratio of bounding box, ncols/ nrows, result floatvector[value]
# bounding box of whole pic, independent from content, trivial case
def test_aspect_ratio():
    
    #case 1
    img = Image((0,0), (8,8), ONEBIT)
    ratio_f = img.aspect_ratio()
    assert ratio_f[0] == 1.0
    
    #case 2
    img = Image((0,0), (3,5), ONEBIT)
    ratio_f = img.aspect_ratio()
    assert abs(ratio_f[0] - 2/3.0) <= eps # must be 0, kinda redundant
    assert abs(ratio_f[0] - 0.6666666) <= eps
    
    #case 3
    img = Image((0,0), (3,16), ONEBIT)
    ratio_f = img.aspect_ratio()
    assert abs(ratio_f[0] - 0.235294118) <= eps
    

# black_area , returns number of black pixel, result floatvector[value]
def test_black_area():
    
    #case 1
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_filled_rect((1,1),(3,5),1)
    black_area_f = img.black_area()
    assert black_area_f[0] == 15.0
    
    #case 2
    img.draw_line((0,3),(7,3),1)
    black_area_f = img.black_area()
    assert black_area_f[0] == 20.0
    
    #case 4
    img.set((7,7),1)
    black_area_f = img.black_area()
    assert black_area_f[0] == 21.0
    
    #case 5
    img.set((3,4),1)
    black_area_f = img.black_area()
    assert black_area_f[0] == 21.0
    
    #case 6
    # would theoretically alert, if interpolation algorithm changed,
    # because no trivial diagonals were drawn
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((1,1),(3,5),1)
    img.draw_line((4,2),(1,6),1)
    black_area_f = img.black_area()
    assert black_area_f[0] == 10.0
    
    
# compactness, long lines low-, circle high-compactness, 
# result floatvector[value]
# size of pic has influence, i.e. 1 lonely px in 3x3 pic has different 
# compactness than 1px in 5x5 pic, empty pic has by definition max 
# compactness (1,79.. e+308), full pic has compactness=0
def test_compactness():
    
    #case 0 complete filled
    img = Image((0,0),(7,7), ONEBIT)
    img.draw_filled_rect((0,0),(7,7),1)
    comp_f = img.compactness()
    assert comp_f[0] == 0.5625
    
    #case 1 rect with white border
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_filled_rect((1,1),(6,6),1)
    comp_f = img.compactness()
    assert abs(comp_f[0] - 0.777777777) <= eps
    
    #case 2 one single pixel
    img = Image((0,0), (7,7), ONEBIT)
    img.set((1,1),1)
    comp_f = img.compactness()
    assert comp_f[0] == 8.0
    
    #case 3 black border
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((0,0),(7,0),1)
    img.draw_line((7,0),(7,7),1)
    img.draw_line((7,7),(0,7),1)
    img.draw_line((0,7),(0,0),1)
    comp_f = img.compactness()
    assert comp_f[0] == 2.0
    
    #case 4 circle
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_circle((3,3),3,1)
    comp_f = img.compactness()
    assert abs(comp_f[0] - 2.647058823529) <= eps
    
    #case 5 horizontal line
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((2,2),(6,2),1)
    comp_f = img.compactness()
    assert abs(comp_f[0] - 3.2) <= eps
    
    #case 6 diagonal line
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((1,2),(2,7),1)
    comp_f = img.compactness()
    assert abs(comp_f[0] - 10/3.0) <= eps
    
   
# moments, result floatvector[X,Y,mom1,mom2,..,mom7]
# X,Y center of gravity normalized on 1.0
# position in relation to surrounding bounding box matters
def test_moments():
    
    #case 1
    img = Image((0,0), (8,8), ONEBIT)
    img.set((4,1),1)
    mom_f = img.moments()
    assert mom_f[0] == 0.5
    assert mom_f[1] == 0.125
    assert mom_f[2] == 0 # single 1px has no moments
    assert mom_f[8] == 0
    
    #case 1b central symmetric rect
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_filled_rect((2,3),(5,4),1)
    mom_f = img.moments()
    assert mom_f[0] == 0.5
    assert mom_f[1] == 0.5
    assert mom_f[2] == 0.15625
    assert mom_f[3] == 0.03125
    assert mom_f[4] == 0 
    assert mom_f[5] == 0
    assert mom_f[6] == 0
    assert mom_f[7] == 0
    assert mom_f[8] == 0
    
    #case 2
    #like the Letter L on the left handside
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((1,2),(1,5),1)
    img.draw_line((1,5),(3,5), 1) 
    mom_f = img.moments()
    assert abs(mom_f[0] - 0.21428571) <= eps
    assert abs(mom_f[1] - 0.57142857) <= eps
    assert abs(mom_f[2] - 0.09722222) <= eps
    assert abs(mom_f[3] - 0.22222222) <= eps
    assert abs(mom_f[4] - 0.08333333) <= eps
    assert abs(mom_f[5] - 0.03402069) <= eps
    assert abs(mom_f[6] + 0.01134023) <= eps
    assert abs(mom_f[7] - 0.02268046) <= eps
    assert abs(mom_f[8] + 0.06804138) <= eps 
    
    #case 3
    img = Image((0,0), (8,8), ONEBIT)
    img.draw_line((1,1),(5,4),1)
    img.draw_circle((5,4),2, 1) 
    mom_f = img.moments()
    assert abs(mom_f[0] - 0.5294117647058) <= eps
    assert abs(mom_f[1] - 0.41911764705882) <= eps
    assert abs(mom_f[2] - 0.142072053734) <= eps
    assert abs(mom_f[4] - 0.057398738041) <= eps
    assert abs(mom_f[7] + 0.0240267688551) <= eps # neg. moment
    
   
# ncols_feature, simply num of cols, result floatvector[value]
def test_ncols_feature():
    
    #case 1
    img = Image((0,0), (7,7), ONEBIT)
    ncols_f = img.ncols_feature()
    assert ncols_f[0] == 8.0
    
    #case 2
    img = Image((0,0), (3,8), ONEBIT)
    ncols_f = img.ncols_feature()
    assert ncols_f[0] == 4.0
    

# nholes, num of white runs (not border) avg, result floatvector[X, Y]
def test_nholes():
    
    #case 1
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_filled_rect((1,1), (3,5), 1)
    nholes_f = img.nholes()
    assert nholes_f[0] == 0.0
    
    #case 2
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_hollow_rect((1,1), (4,5), 1) 
    img.draw_hollow_rect((2,3), (6,7), 1) 
    img.draw_line((3,7),(5,7),0) #white line
    
    nholes_f = img.nholes()
    assert nholes_f[0] == 0.375 # 3 holes per 8 lines
    assert nholes_f[1] == 0.75 # 6 holes per 8 lines
    
    #case 2b // same pic double size to test scale-invariance
    img = img.scale(2.0,0)
    
    nholes_f = img.nholes()
    assert abs(nholes_f[0] - 0.375) <= eps 
    assert abs(nholes_f[1] - 0.75) <= eps
    
    
# nholes_extended, divides pic in 4 parts and does nholes analysis on each of it
# , result floatvector[X, Y]
def test_nholes_extended():
    
    #case 1
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_filled_rect((1,1), (3,5), 1)
    nholes_ex_f = img.nholes_extended()
    assert nholes_ex_f[0] == 0.0
    
    #case 2
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_hollow_rect((1,4), (3,6), 1)
    img.draw_hollow_rect((3,2), (5,5), 1)  
    img.draw_hollow_rect((1,0), (3,2), 1)      
    img.draw_line((1,0),(3,0),0) #white line
    img.set(Point(3,3),0) #white Point
    img.set(Point(7,5),1)
    nholes_ex_f = img.nholes_extended()
    #vertical
    assert nholes_ex_f[0] == 0.5    
    assert nholes_ex_f[1] == 1.5    
    assert nholes_ex_f[2] == 0.5  
    assert nholes_ex_f[3] == 0.0    
    #horizontal
    assert nholes_ex_f[4] == 0.5  
    assert nholes_ex_f[5] == 0.0    
    assert nholes_ex_f[6] == 1.5  
    assert nholes_ex_f[7] == 0.0  
     
    
# nrows_feature, simply num of rows, result floatvector[value]
def test_nrows_feature():
    
    #case 1
    img = Image((0,0), (7,7), ONEBIT)
    nrows_f = img.nrows_feature()
    assert nrows_f[0] == 8.0
    
    #case 2
    img = Image((0,0), (7,5), ONEBIT)
    nrows_f = img.nrows_feature()
    assert nrows_f[0] == 6.0   
    
    
# skeleton_features, thins out a structure to a skeleton and runs several tests
# on it, returns floatvector[numOf 4connected pxs ~ X, numOf 3connectd pxs ~ T,
# avgOf bendpoints, numOf endPoints, numOf x-axis crossing pxs, numOf y-axis 
# crossing pxs] 
# x- and y-axis through centre of masse
def test_skeleton_features():
    
    # capital Letter A with unsteadiness like hand-written
    img = Image((0,0), (49,49), ONEBIT)
    img.draw_filled_rect((3,41),(13,45),1) # left base of letter A
    img.draw_filled_rect((31,42),(42,46),1) # right base of letter A
    img.draw_line((5,40),(12,40),1)
    img.draw_filled_rect((6,38),(13,39),1)
    img.draw_line((7,37),(14,37),1)
    img.draw_line((8,36),(27,36),1)
    img.draw_filled_rect((9,34),(28,35),1)
    img.draw_filled_rect((10,32),(35,33),1)
    img.draw_line((11,31),(34,31),1)
    img.draw_line((11,30),(18,30),1)
    img.draw_line((12,29),(18,29),1)
    img.draw_line((12,28),(19,28),1)
    img.draw_line((13,27),(20,27),1)
    img.draw_filled_rect((14,25),(20,26),1)
    img.draw_line((15,24),(21,24),1)
    img.draw_line((16,23),(21,23),1)
    img.draw_filled_rect((16,21),(22,22),1)
    img.draw_filled_rect((17,19),(23,20),1)
    img.draw_filled_rect((18,17),(23,18),1)
    img.draw_filled_rect((19,13),(29,16),1)
    img.draw_filled_rect((20,10),(28,12),1)
    img.draw_filled_rect((21,8),(27,9),1)
    img.draw_line((21,7),(26,7),1)
    img.draw_filled_rect((22,5),(26,6),1)
    img.draw_filled_rect((23,16),(30,17),1)
    img.draw_filled_rect((25,18),(30,19),1)
    img.draw_line((25,20),(31,20),1)
    img.draw_filled_rect((26,21),(31,23),1)
    img.draw_filled_rect((27,24),(32,27),1)
    img.draw_filled_rect((28,27),(33,28),1)
    img.draw_line((27,29),(34,29),1)
    img.draw_line((24,30),(34,30),1)
    img.draw_line((29,34),(36,34),1)
    img.draw_line((30,35),(36,35),1)
    img.draw_line((31,36),(36,36),1)
    img.draw_line((31,37),(37,37),1)
    img.draw_line((32,38),(37,38),1)
    img.draw_filled_rect((32,39),(38,40),1)
    img.draw_line((33,41),(40,41),1)
    
    skel_f = img.skeleton_features()
    assert skel_f[0] == 2.0    # num of X-Joins (4 connected pxs)
    assert skel_f[1] == 10.0   # num of T-Joins (3 connected pxs) ?!?!?
    assert abs(skel_f[2] - 0.40659340) <= eps # avg of bend points
    assert skel_f[3] == 4.0    # num of endpoints
    assert skel_f[4] == 2.0    # num of x-crossings through center
    assert skel_f[5] == 2.0    # num og y-crossings through center
    

# top_bottom, simply detects the first and last row in which black Pixel appear,
# returns floatvector[firstrow/nrows, lastrow/nrows]
def test_top_bottom():
    
    #like the Letter L on the left handside
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((1,2),(1,5),1)
    img.draw_line((1,5),(3,5), 1)    
    
    top_bottom_f = img.top_bottom()
    assert top_bottom_f[0] == 0.25
    assert top_bottom_f[1] == 0.625
 
    
# volume, numOf black pxs (black_area) / (ncols*nrows) (~bounding box), returns
# floatvector[value]
def test_volume():
    #like the Letter L on the left handside
    img = Image((0,0), (7,7), ONEBIT)
    img.draw_line((1,2),(1,5),1)
    img.draw_line((1,5),(3,5), 1)    
    
    vol_f = img.volume()
    assert vol_f[0] == 0.09375
  
    
# volume16regions, divides the pic in 16 regions and calculate the volume of each
# region seperatly (black_area/(region.cols*region.rows), returnsfloatvector[ 
# first column from top to bottom, 2nd column ... ] 
def test_volume_16_regions():
    
    img = Image((0,0),(7,7), ONEBIT)
    img.draw_hollow_rect((1,1),(6,4),1)
    img.draw_hollow_rect((2,4),(7,7),1)
    
    vol16_f = img.volume16regions()
    assert vol16_f[0] == 0.25 
    assert vol16_f[1] == 0.5
    assert vol16_f[2] == 0.25
    assert vol16_f[3] == 0
    assert vol16_f[4] == 0.5
    assert vol16_f[5] == 0
    assert vol16_f[6] == 0.75
    assert vol16_f[7] == 0.75
    assert vol16_f[8] == 0.5
    assert vol16_f[9] == 0
    assert vol16_f[10] == 0.5
    assert vol16_f[11] == 0.5
    assert vol16_f[12] == 0.25
    assert vol16_f[13] == 0.5
    assert vol16_f[14] == 0.75
    assert vol16_f[15] == 0.75

    img = Image((0,0),(0,0),ONEBIT)
    img.set((0,0),1)
    vol16_f = img.volume16regions()
    for i in range(15):
        assert vol16_f[i] == 1.0
    
    
# volume64regions, divides the pic in 64 regions and calculate the volume of each
# region seperatly (black_area/(region.cols*region.rows), returnsfloatvector[ 
# first column from top to bottom, 2nd column ... ] 
def test_volume_64_regions():
    
    img = Image((0,0),(15,15),ONEBIT)
    img.draw_hollow_rect((0,0),(15,15),1)
    img.draw_filled_rect((7,7),(8,8),1)
    
    vol64_f = img.volume64regions()
    assert vol64_f[0] == 0.75            #ul-region
    assert vol64_f[1] == 0.5             #1st col, 2nd row
    assert vol64_f[7] == 0.75            #ll-region
    assert vol64_f[8] == 0.5             #2nd column, 1st row
    assert vol64_f[9] == 0               #2nd column, 2nd row
    assert vol64_f[27] == 0.25           #ul-region from center
    assert vol64_f[63] == 0.75           #lr-region
    
    img = Image((0,0),(0,0),ONEBIT)
    img.set((0,0),1)
    vol64_f = img.volume64regions()
    for i in range(63):
        assert vol64_f[i] == 1.0
        
# zernike_moments, more complex feature extraction, the shape of the origin
# can be reconstructed outta the first few ZMs. The higher order moments 
# describe the shape even better up to the perfect copy for order n to 
# infinity.
# returnsfloatvector[magnitudes of NZM20, NZM22, ... , NZM66] 
        
def test_zernike_moments():
    
    #big capital-L
    img = Image((0,0),(50,50),ONEBIT)
    img.draw_line((5,5),(5,35),3)
    img.draw_line((3,35),(20,35),1)
    
    ZM_f0 = img.zernike_moments()
    assert abs(ZM_f0[0] -  0.4041497) <= eps
    assert abs(ZM_f0[1] -  0.2138501) <= eps
    assert abs(ZM_f0[2] -  0.2063463) <= eps
    assert abs(ZM_f0[3] -  0.1483378) <= eps
    assert abs(ZM_f0[4] -  0.1623111) <= eps
    assert abs(ZM_f0[5] -  0.3384352) <= eps
    assert abs(ZM_f0[6] -  0.1148943) <= eps
    assert abs(ZM_f0[7] -  0.2504674) <= eps
    assert abs(ZM_f0[8] -  0.0819277) <= eps
    assert abs(ZM_f0[9] -  0.1827993) <= eps
    assert abs(ZM_f0[10] - 0.0524176) <= eps
    assert abs(ZM_f0[11] - 0.4059857) <= eps
    assert abs(ZM_f0[12] - 0.0920007) <= eps
    assert abs(ZM_f0[13] - 0.1367013) <= eps

    # test rotation-invariance
    img = img.rotate(90.0, 1)
    ZM_f = img.zernike_moments()
    assert abs(ZM_f[0] -  ZM_f0[0]) <= eps
    assert abs(ZM_f[1] -  ZM_f0[1]) <= eps
    assert abs(ZM_f[2] -  ZM_f0[2]) <= eps
    assert abs(ZM_f[3] -  ZM_f0[3]) <= eps
    assert abs(ZM_f[4] -  ZM_f0[4]) <= eps
    assert abs(ZM_f[5] -  ZM_f0[5]) <= eps
    assert abs(ZM_f[6] -  ZM_f0[6]) <= eps
    assert abs(ZM_f[7] -  ZM_f0[7]) <= eps
    assert abs(ZM_f[8] -  ZM_f0[8]) <= eps
    assert abs(ZM_f[9] -  ZM_f0[9]) <= eps
    assert abs(ZM_f[10] - ZM_f0[10]) <= eps
    assert abs(ZM_f[11] - ZM_f0[11]) <= eps
    assert abs(ZM_f[12] - ZM_f0[12]) <= eps
    assert abs(ZM_f[13] - ZM_f0[13]) <= eps
    
    # test mirror-invariance
    img.mirror_horizontal()
    ZM_f = img.zernike_moments()
    assert abs(ZM_f[0] -  ZM_f0[0]) <= eps
    assert abs(ZM_f[1] -  ZM_f0[1]) <= eps
    assert abs(ZM_f[2] -  ZM_f0[2]) <= eps
    assert abs(ZM_f[3] -  ZM_f0[3]) <= eps
    assert abs(ZM_f[4] -  ZM_f0[4]) <= eps
    assert abs(ZM_f[5] -  ZM_f0[5]) <= eps
    assert abs(ZM_f[6] -  ZM_f0[6]) <= eps
    assert abs(ZM_f[7] -  ZM_f0[7]) <= eps
    assert abs(ZM_f[8] -  ZM_f0[8]) <= eps
    assert abs(ZM_f[9] -  ZM_f0[9]) <= eps
    assert abs(ZM_f[10] - ZM_f0[10]) <= eps
    assert abs(ZM_f[11] - ZM_f0[11]) <= eps
    assert abs(ZM_f[12] - ZM_f0[12]) <= eps
    assert abs(ZM_f[13] - ZM_f0[13]) <= eps
    
    # test scale-invariance
    img = img.scale(3, 1)
    ZM_f = img.zernike_moments()
    assert abs(ZM_f[0] -  ZM_f0[0]) <= 0.1
    assert abs(ZM_f[1] -  ZM_f0[1]) <= 0.1
    assert abs(ZM_f[2] -  ZM_f0[2]) <= 0.1
    assert abs(ZM_f[3] -  ZM_f0[3]) <= 0.1
    assert abs(ZM_f[4] -  ZM_f0[4]) <= 0.1
    assert abs(ZM_f[5] -  ZM_f0[5]) <= 0.1
    assert abs(ZM_f[6] -  ZM_f0[6]) <= 0.1
    assert abs(ZM_f[7] -  ZM_f0[7]) <= 0.1
    assert abs(ZM_f[8] -  ZM_f0[8]) <= 0.1
    assert abs(ZM_f[9] -  ZM_f0[9]) <= 0.1
    assert abs(ZM_f[10] - ZM_f0[10]) <= 0.1
    assert abs(ZM_f[11] - ZM_f0[11]) <= 0.1
    assert abs(ZM_f[12] - ZM_f0[12]) <= 0.1
    assert abs(ZM_f[13] - ZM_f0[13]) <= 0.1



import py.test

from gamera.core import *
init_gamera()

#
# Tests for converting color image to labeld CCs
#

def test_color_to_ccs():

    # when more colors than labels: exception
    def _too_many_colors():
        img = load_image("data/too_many_colors.png")
        labeled = img.colors_to_labels()
    py.test.raises(Exception, _too_many_colors)

    # test labeling
    img = Image((0,0), (40,40), RGB)
    img.draw_filled_rect((3,3),(5,5),RGBPixel(255,0,0))
    img.draw_filled_rect((10,3),(15,35),RGBPixel(0,255,0))
    img.draw_filled_rect((5,20),(30,25),RGBPixel(0,0,255))
    labeled = img.colors_to_labels({\
                RGBPixel(255,0,0):6,\
                RGBPixel(0,255,0):2,\
                RGBPixel(0,0,255):7,\
              })
    assert 6 == labeled.get((4,4))
    assert 2 == labeled.get((12,18))
    assert 7 == labeled.get((12,22))

    # test conversion to CCs
    ccs = labeled.ccs_from_labeled_image()
    for c in ccs:
        if c.label == 6:
            assert [(3,3),(5,5)] == [c.ul, c.lr]
        if c.label == 2:
            assert [(10,3),(15,35)] == [c.ul, c.lr]
        if c.label == 7:
            assert [(5,20),(30,25)] == [c.ul, c.lr]



import sys

def my_application():
   global ccs

   # Load the image
   image = load_image(sys.argv[-1])
   onebit = image.to_onebit()
   # Get the connected components from the image
   ccs = onebit.cc_analysis()
   # Classify the cc's
   classifier.classify_list_automatic(ccs)

   # Display the ccs to show their classification
   display_multi(ccs)

# Import the Gamera core and initialise it
from gamera.core import *
init_gamera()

# Import the classifier module
from gamera import knn
# Create a new classifier
classifier = knn.kNNInteractive()
# Load some training data
classifier.from_xml_filename("training.xml")

# Import the Gamera GUI and start it   
from gamera.gui import gui
gui.run(my_application)




#
# Copyright (C) 2005 Alex Cobb
#               2015 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Data transfer between Gamera and the Python Imaging Library.
"""

from gamera.plugin import *
from gamera import config

try:
    #import Image as PIL
    from PIL import Image as PIL
except ImportError:
    try:
        verbose = config.get("verbosity_level")
    except Exception:
        verbose = 0
    if verbose:
        print ('Info: Python Imaging Library module Image '
               'could not be imported')
else:

    _modes = {RGB       : 'RGB',
              GREYSCALE : 'L'}
    _inverse_modes = {'RGB': RGB,
                      'L'  : GREYSCALE}
    ## also boolean mode '1' --- I think this is 8 bits per pixel

    class from_pil(PluginFunction):
        """
        Instantiates a Gamera image from a Python Imaging Library
        image *image*.

        Only RGB or 8-bit greyscale mode PIL images are supported.
        Requires a copying operation;  may fail for very large images.

        This can, e.g., be used to read images from file formats not
        directly supported by gamera, as JPEG images:

        .. code:: Python

          # Beware: name "Image" is already used in Gamera!
          import Image as Pil
          from pil_io import

          # read a JPEG image and convert it to a Gamera image
          pilimg = Pil.open("image.jpg")
          img = from_pil(pilimg)
        """
        self_type = None
        return_type = ImageType([GREYSCALE, RGB, FLOAT])
        args = Args([Class("image")])
        pure_python = True

        def __call__(image, offset=None):
            from gamera.plugins import _string_io
            from gamera.core import Dim, Point
            typecode = image.mode
            if offset is None:
                offset = Point(0, 0)
            if _inverse_modes.has_key(typecode):
                pixel_type = _inverse_modes[typecode]
            else:
                raise ValueError("Only RGB and 8-bit Greyscale 'L' PIL image modes are supported.")
            try:
                return _string_io._from_raw_string(
                    offset,
                    Dim(image.size[0], image.size[1]),
                    pixel_type, DENSE,
                    image.tobytes())
            except Exception:
                # for compatibility with pil 1.1.7 and earlier
                return _string_io._from_raw_string(
                    offset,
                    Dim(image.size[0], image.size[1]),
                    pixel_type, DENSE,
                    image.tostring())
        __call__ = staticmethod(__call__)

    class to_pil(PluginFunction):
        """
        Returns a Python Imaging Library image containing a copy of
        image's data.

        Only RGB and Greyscale images are supported.
        May fail for very large images.
        """
        self_type = ImageType([RGB, GREYSCALE])
        return_type = Class("pil_image")
        def __call__(image):
            from gamera.plugins import _string_io
            pixel_type = image.data.pixel_type
            if _modes.has_key(pixel_type):
                mode = _modes[pixel_type]
            else:
                raise ValueError("Only RGB and GREYSCALE Images are supported.")
            size = (image.ncols, image.nrows)
            try:
                return PIL.frombytes(mode, size,
                                     _string_io._to_raw_string(image))
            except Exception:
                # for compatibility with pil 1.1.7 and earlier
                return PIL.fromstring(mode, size,
                                     _string_io._to_raw_string(image))
        __call__ = staticmethod(__call__)

        def __doc_example1__(images):
            image = images[RGB]
            array = image.to_pil()
            image0 = from_pil(array)
            return [image, image0]
        def __doc_example2__(images):
            image = images[GREYSCALE]
            array = image.to_pil()
            image0 = from_pil(array)
            return [image, image0]
        doc_examples = [__doc_example1__,
                        __doc_example2__]

    class PilIOModule(PluginModule):
        category = "ExternalLibraries/PIL"
        author = "Alex Cobb"
        functions = [from_pil, to_pil]
        pure_python = True
        url = ('http://www.oeb.harvard.edu/faculty/holbrook/'
               'people/alex/Website/alex.htm')
    module = PilIOModule()

    from_pil = from_pil()



#
# Copyright (C) 2012 Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx

from gamera.gui.gaoptimizer.SettingsPanel import *
from gamera.gui.gaoptimizer.StatusPanel import *
from gamera.gui.gaoptimizer.ResultPanel import *

from gamera import knnga
from gamera.gui import gui_util

#-------------------------------------------------------------------------------
class OptimizerFrame(wx.Frame):
#-------------------------------------------------------------------------------
    #---------------------------------------------------------------------------
    def __init__(self, parent, id, title):
    #---------------------------------------------------------------------------
        wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition, \
                          wx.Size(800,800))

        self.classifier = None
        self.settingsFilename = None

        # Status Bar
        statusBar = self.CreateStatusBar()

        # Menu    
        menuBar = wx.MenuBar()
        menuBar.Append(self.CreateFileMenu(), "&File")
        menuBar.Append(self.CreateClassifierMenu(), "&Classifier")
        self.SetMenuBar(menuBar)

        # Notebook with tabs
        self.notebook = wx.Notebook(self, -1)
        self.settingsPanel = SettingsPanel(self.notebook, -1, self)
        self.statusPanel = StatusPanel(self.notebook, -1, self)
        self.resultPanel = ResultPanel(self.notebook, -1, self)
        self.notebook.AddPage(self.settingsPanel, "Settings")
        self.notebook.AddPage(self.statusPanel, "Status")
        self.notebook.AddPage(self.resultPanel, "Results")

        # timer for status updates
        self.timer = wx.Timer(self, -1)
        self.Bind(wx.EVT_TIMER, self.OnTimerUpdate, self.timer)

    #---------------------------------------------------------------------------
    def CreateFileMenu(self):
    #---------------------------------------------------------------------------
        file_menu = wx.Menu()

        # load menu entries
        open_xml = file_menu.Append(wx.NewId(), "Open from &XML", \
            "classifier data from an existing XML file")
        open_classifier = file_menu.Append(wx.NewId(), "Open existing &classifier",
            "Use an existing classiefier from gamera")
        self.Bind(wx.EVT_MENU, self.OnOpenXML, open_xml)
        self.Bind(wx.EVT_MENU, self.OnOpenClassifier, open_classifier)

        file_menu.AppendSeparator()

        # save menu entries
        save_xml = file_menu.Append(wx.NewId(), "&Save to XML", \
            "Save the current settings to XML")
        save_xml.Enable(False)
        save_xml_as = file_menu.Append(wx.NewId(), "Save to XML &as...", \
            "Save the current settings to XML")
        save_xml_as.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnSaveXML, save_xml)
        self.Bind(wx.EVT_MENU, self.OnSaveXMLAs, save_xml_as)

        file_menu.AppendSeparator()

        # exit menu entry
        exit = file_menu.Append(wx.NewId(), "&Exit", "Exit the application")
        self.Bind(wx.EVT_MENU, self.OnExit, exit)
        self.Bind(wx.EVT_CLOSE, self.OnExit)

        return file_menu

    #---------------------------------------------------------------------------
    def CreateClassifierMenu(self):
    #---------------------------------------------------------------------------
        classifier_menu = wx.Menu()

        # copy existing classifier
        copy_classifier = classifier_menu.Append(wx.NewId(), "&Copy classifier", \
            "Create a copy from classifier and use this")
        copy_classifier.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnCopyClassifier, copy_classifier)

        # insert classifier into gamera gui
        into_gui = classifier_menu.Append(wx.NewId(), "Classifier to &Gamera-GUI", \
            "Insert classifier into gamera gui window")
        into_gui.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnIntoGameraGUI, into_gui)

        # load kNN settings xml (selections, weights, num_k, distance type)
        load_settings = classifier_menu.Append(wx.NewId(), "Load &settings from XML", \
            "Load kNN settings XML into current classifier")
        load_settings.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnLoadSettings, load_settings)

        classifier_menu.AppendSeparator()

        # feature set
        change_features = classifier_menu.Append(wx.NewId(), "Change &feature set", \
            "Set the feature set for the classifier")
        change_features.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnChangeFeatures, change_features)

        # reset selection settings
        reset_selections = classifier_menu.Append(wx.NewId(), "Reset se&lection", \
            "Reset the feature selection values from classifier")
        reset_selections.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnResetSelections, reset_selections)

        # reset weight settings
        reset_weights = classifier_menu.Append(wx.NewId(), "Reset &weights", \
            "Reset the feature weights values from classifier")
        reset_weights.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnResetWeights, reset_weights)

        return classifier_menu

    #---------------------------------------------------------------------------
    def EnableControls(self, flag=True):
    #---------------------------------------------------------------------------
        menuBar = self.GetMenuBar()

        id = menuBar.FindMenuItem("&File", "&Save to XML")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&File", "Save to XML &as...")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "&Copy classifier")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "Classifier to &Gamera-GUI")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "Load &settings from XML")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "Change &feature set")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "Reset se&lection")
        menuBar.FindItemById(id).Enable(flag)

        id = menuBar.FindMenuItem("&Classifier", "Reset &weights")
        menuBar.FindItemById(id).Enable(flag)

        self.settingsPanel.startButton.Enable()

    #---------------------------------------------------------------------------
    def OnTimerUpdate(self, event):
    #---------------------------------------------------------------------------
        self.UpdatePanels()

    #---------------------------------------------------------------------------
    def UpdatePanels(self):
    #---------------------------------------------------------------------------
        self.statusPanel.Update()
        self.resultPanel.Update()

    #---------------------------------------------------------------------------
    def SetClassifiers(self):
    #---------------------------------------------------------------------------
        self.settingsPanel.classifier = self.classifier
        self.statusPanel.classifier = self.classifier
        self.resultPanel.classifier = self.classifier

    #---------------------------------------------------------------------------
    def CopyClassifier(self, classifier):
    #---------------------------------------------------------------------------
        from gamera import knn

        copiedDatabase = []
        for img in self.classifier.get_database():
            imgCopy = img.image_copy()
            imgCopy.id_name = img.id_name
            imgCopy.classification_state = img.classification_state
            copiedDatabase.append(imgCopy)

        copiedClassifier = knn.kNNNonInteractive(
            copiedDatabase,
            self.classifier.features,
            self.classifier._perform_splits,
            self.classifier.num_k,
            self.classifier.normalize)
        copiedClassifier.set_weights(self.classifier.get_weights())
        copiedClassifier.set_selections(self.classifier.get_selections())

        return copiedClassifier

    #---------------------------------------------------------------------------
    def OnOpenXML(self, event):
    #---------------------------------------------------------------------------
        from gamera import gamera_xml
        from gamera import knn

        filename = gui_util.open_file_dialog(self, gamera_xml.extensions)
        if filename == None:
            gui_util.message("Can't open classifier-xml file")
            return

        classifier = knn.kNNNonInteractive(filename)
        self.classifier = classifier
        self.SetClassifiers()

        self.UpdatePanels()
        self.EnableControls(True)

    #---------------------------------------------------------------------------
    def OnOpenClassifier(self, event):
    #---------------------------------------------------------------------------
        from gamera.args import Args, Class
        from gamera.gui import gui
        from gamera import knn

        dialog = Args(
            [Class("Classifier", knn._kNNBase)],
             name="Select an existing classifier...")
        results = dialog.show(
            self, gui.main_win.shell.locals,
            docstring="""Choose an already created classifier for optimization.""")

        if results == None:
            return

        self.classifier = results[0]

        if isinstance(self.classifier, knn.kNNInteractive):
            gui_util.message("Given interactive classifier will be converted"\
                " to noninteractive classifier for optimization (internal)")
            self.classifier = self.CopyClassifier(self.classifier)

        self.SetClassifiers()
        self.UpdatePanels()
        self.EnableControls(True)

    #---------------------------------------------------------------------------
    def OnSaveXML(self, event):
    #---------------------------------------------------------------------------
        if self.classifier == None:
            gui_util.message("No classifier loaded")
            return

        if self.settingsFilename == None:
            self.OnSaveXMLAs(event)

        # save the current settings into file
        wx.BeginBusyCursor()
        try:
            self.classifier.save_settings(self.settingsFilename)
        finally:
            wx.EndBusyCursor()

    #---------------------------------------------------------------------------
    def OnSaveXMLAs(self, event):
    #---------------------------------------------------------------------------
        if self.classifier == None:
            gui_util.message("No classifier loaded")
            return

        filename = gui_util.save_file_dialog(self)
        self.settingsFilename = filename

        # save the current settings into file
        wx.BeginBusyCursor()
        try:
            self.classifier.save_settings(self.settingsFilename)
        finally:
            wx.EndBusyCursor()

    #---------------------------------------------------------------------------
    def OnExit(self, event):
    #---------------------------------------------------------------------------
        response = wx.MessageDialog(self, "Are you sure you want to quit the "\
            "optimization?\n(Make sure you have saved your results)", \
            "Exit?", wx.YES_NO).ShowModal()

        if response == wx.ID_NO:
            return False

        if self.settingsPanel.stopCalculation():
            self.Destroy()
        else:
            return

    #---------------------------------------------------------------------------
    def OnCopyClassifier(self, event):
    #---------------------------------------------------------------------------
        from gamera import knn

        if self.classifier == None:
            gui.message("No classifier loaded")
            return

        if self.classifier.is_interactive():
            self.classifier = self.classifier.noninteractive_copy()
        else:
            self.classifier = self.CopyClassifier(self.classifier)

    #---------------------------------------------------------------------------
    def OnIntoGameraGUI(self, event):
    #---------------------------------------------------------------------------
        from gamera import knn
        from gamera.gui import gui
        from gamera.gui import var_name

        name = var_name.get("classifier", gui.main_win.shell.locals)

        if name == None:
            gui_util.message("No valid variable name for classifier entered")
            return

        copiedClassifier = self.CopyClassifier(self.classifier)

        gui.main_win.shell.locals[name] = copiedClassifier
        gui.main_win.shell.run('\n')

    #---------------------------------------------------------------------------
    def OnLoadSettings(self, event):
    #---------------------------------------------------------------------------
        from gamera import gamera_xml

        if self.classifier == None:
            gui.message("No classifier loaded")
            return

        filename = gui_util.open_file_dialog(self, gamera_xml.extensions)
        if filename == None:
            gui_util.message("Can't open classifier settings xml file")
            return

        self.classifier.load_settings(filename)

        self.UpdatePanels()

    #---------------------------------------------------------------------------
    def OnChangeFeatures(self, event):
    #---------------------------------------------------------------------------
        from gamera import plugin
        from gamera.core import ONEBIT
        from gamera.core import ImageBase
        from gamera.args import Args, Check

        if self.classifier == None:
            gui.message("No classifier loaded")
            return

        allFeatures = [x[0] for x in plugin.methods_flat_category("Features", ONEBIT)]
        allFeatures.sort()

        existingFeatures = [x[0] for x in ImageBase.get_feature_functions(self.classifier.features)[0]]

        featureControls = []
        for f in allFeatures:
            featureControls.append(Check('', f, default=(f in existingFeatures)))

        dialog = Args(featureControls, name = "Feature selection",
            title="Select the features you want to use for optimization")

        result = dialog.show(self)

        if result == None:
            gui_util.message("No change applied")
            return

        selectedFeatures = [name for check, name in zip(result, allFeatures) if check]
        self.classifier.change_feature_set(selectedFeatures)

        self.UpdatePanels()

    #---------------------------------------------------------------------------
    def OnResetSelections(self, event):
    #---------------------------------------------------------------------------
        if self.classifier == None:
            gui.message("No classifier loaded")
            return

        selections = self.classifier.get_selections()

        for sIndex in range(len(selections)):
            selections[sIndex] = 1

        self.classifier.set_selections(selections)

        self.SetClassifiers()
        self.UpdatePanels()

    #---------------------------------------------------------------------------
    def OnResetWeights(self, event):
    #---------------------------------------------------------------------------
        if self.classifier == None:
            gui.message("No classifier loaded")
            return

        weights = self.classifier.get_weights()

        for wIndex in range(len(weights)):
            weights[wIndex] = 1.0

        self.classifier.set_weights(weights)

        self.SetClassifiers()
        self.UpdatePanels()



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

ONEBIT = 0
GREYSCALE = 1
GREY16 = 2
RGB = 3
FLOAT = 4
COMPLEX = 5
ALL = [ONEBIT, GREYSCALE, GREY16, RGB, FLOAT, COMPLEX]
NON_COMPLEX_TYPES = ALL[:]
NON_COMPLEX_TYPES.remove(COMPLEX)
NONIMAGE = -1

DENSE = 0
RLE = 1




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx
import weakref              # Python standard library
from gamera.core import *   # Gamera-specific
from gamera import util, plugin
from gamera.gui import var_name, gui_util
from gamera.args import *

EXECUTE_MODE = 0
HELP_MODE = 1

shell = None
def set_shell(sh):
   global shell
   shell = sh

def set_shell_frame(sf):
   global shell_frame
   shell_frame = sf

######################################################################

_members_for_menu = ('pixel_type_name',
                     'storage_format_name',
                     'ul_x', 'ul_y', 'nrows', 'ncols',
                     'resolution', 'memory_size', 'label',
                     'classification_state', 'properties')
def members_for_menu(self):
   return ["%s: %s" % (x, getattr(self, x))
           for x in _members_for_menu
           if hasattr(self, x)]

def methods_for_menu(self):
   return plugin.plugin_methods[self.data.pixel_type]

class ImageMenu:
   _base_method_id = 10003

   def __init__(self, parent, x, y, images_, name_=None, shell_=None,
                mode=EXECUTE_MODE, extra_methods={}):
      self.shell = shell_
      self.locals = locals
      self.mode = mode
      self.parent = parent
      if not util.is_sequence(images_):
         self.images = [images_]
      else:
         self.images = images_
      if not util.is_homogeneous_image_list(self.images):
         gui_util.message("All selected images are not of the same type.")
      self.image_name = name_

      self._menu_ids = []
      members = members_for_menu(self.images[0])
      methods = methods_for_menu(self.images[0])
      menu = self.create_menu(
        members, methods,
        self.images[0].data.pixel_type,
        self.images[0].pixel_type_name,
        extra_methods)
      self.did_something = 0
      self.parent.PopupMenu(menu, wx.Point(x, y))
      for i in self._menu_ids:
         self.parent.Disconnect(i)
      menu.Destroy()

   def get_id(self):
      id = wx.NewId()
      self._menu_ids.append(id)
      return id

##   def __del__(self):
##     print 'ImageMenu deleted'

   # Given a list of variables and methods, put it all together
   def create_menu(self, members, methods, type, type_name, extra_methods):
      menu = wx.Menu()
      # Top line
      if self.mode == HELP_MODE:
         menu.Append(0, "Help")
      id = self.get_id()
      menu.Append(id, type_name + " Image")
      menu.AppendSeparator()
      id = self.get_id()
      menu.Append(id, "new reference")
      wx.EVT_MENU(self.parent, id, self.OnCreateReference)
      id = self.get_id()
      menu.Append(id, "new copy")
      wx.EVT_MENU(self.parent, id, self.OnCreateCopy)
      id = self.get_id()
      menu.Append(id, "delete image")
      wx.EVT_MENU(self.parent, id, self.OnDeleteImage)
      menu.AppendSeparator()

      info_menu = wx.Menu()
      id = self.get_id()
      menu.AppendMenu(id, "Info", info_menu)
      # Variables
      for member in members:
         id = self.get_id()
         info_menu.Append(id, member)

      # Methods
      menu.AppendSeparator()
      self.functions = {}
      menu = self.create_methods(methods, menu)

      # Extra methods
      if len(extra_methods):
         menu.AppendSeparator()
         menu = self.create_extra_methods(extra_methods, menu)

      return menu

   def create_methods(self, methods, menu):
      items = methods.items()
      items.sort()
      for key, val in items:
         if type(val) == dict:
            item = self.create_methods(val, wx.Menu())
            id = self.get_id()
            menu.AppendMenu(id, key, item)
         else:
            id = self.get_id()
            menu.Append(id, key)
            wx.EVT_MENU(self.parent, id, self.OnPopupFunction)
            self.functions[id] = val
      return menu

   def create_extra_methods(self, methods, menu):
      items = methods.items()
      items.sort()
      for key, val in items:
         if type(val) == dict:
            id = self.get_id()
            item = self.create_extra_methods(val, wx.Menu())
            menu.AppendMenu(id, key, item)
         else:
            id = self.get_id()
            menu.Append(id, key)
            wx.EVT_MENU(self.parent, id, val)
      return menu

   def get_shell(self):
      if self.shell:
         return self.shell
      return shell

   # Creates a new top-level reference to the image
   def OnCreateReference(self, event):
      sh = self.get_shell()
      name = var_name.get("ref", sh.locals)
      if name:
         if len(self.images) == 1:
            if isinstance(self.images[0], weakref.ProxyTypes):
               sh.locals[name] = self.images[0].image_copy()
            else:
               sh.locals[name] = self.images[0]
         else:
            sh.locals[name] = []
            for image in self.images:
               if isinstance(image, weakref.ProxyTypes):
                  sh.locals[name].append(image.image_copy())
               else:
                  sh.locals[name].append(image)
         sh.Update()
      sh.update()

   def OnCreateCopy(self, event):
      sh = self.get_shell()
      name = var_name.get("copy", sh.locals)
      if name:
         if len(self.images) == 1:
            sh.locals[name] = self.images[0].image_copy()
         else:
            sh.locals[name] = []
            for i in range(len(self.images)):
               sh.locals[name].append(self.images[i].image_copy())
         sh.Update()
      sh.update()

   def OnDeleteImage(self,event):
       sh = self.get_shell()
       sh.run("del " + self.image_name)
       sh.Update()

   def get_function_call(self, sh, function):
      # determine if the function requires an argument gui
      if function.args.list in ('', None, (), []):
         # if not, we can just use empty parentheses
         return function.__name__ + "()"
      # else, display the argument gui and use what it returns
      result = function.args.show(
         self.parent, sh.GetLocals(),
         function.__name__, docstring=function.__call__.__doc__)
      return result

   def get_result_name(self, function, dict):
      if function.return_type not in ('', None):
         if function.return_type.name != None:
            name = function.return_type.name
         else:
            name = function.__name__
         return var_name.get(name, dict)
      return ''

   def OnPopupFunction(self, event):
      sh = self.get_shell()
      function = self.functions[event.GetId()]
      if self.images:
         if self.mode == HELP_MODE:
            sh.run("help('" + function.__name__ + "')")
         elif self.mode == EXECUTE_MODE:
            func_call = self.get_function_call(sh, function)
            if func_call == None:  # User pressed cancel, so bail
               return
            result_name = self.get_result_name(function, sh.locals)
            if result_name is None: return
            # If there is no image name, we have to run the code locally (i.e.
            # not in the shell)
            wx.BeginBusyCursor()
            try:
               if self.image_name is None:
                  self._run_locally(sh, result_name, func_call)
               else:
                  self._run_in_shell(sh, result_name, func_call)
            finally:
               wx.EndBusyCursor()
      self.did_something = 1
      sh.update()

   def _run_locally(self, sh, result_name, func_call):
      namespace = {}
      namespace.update(sh.locals)
      if len(self.images) == 1:
         namespace['image'] = self.images[0]
         source = 'image.%s' % func_call
         result = eval(source, namespace)
         if result_name != '':
            sh.locals[result_name] = result
      else:
         namespace['images'] = self.images
         if result_name != '':
            sh.locals[result_name] = []
         progress = util.ProgressFactory(
           'Processing images...', len(self.images))
         try:
            for i in range(len(self.images)):
               source = "images[%d].%s" % (i, func_call)
               result = eval(source, namespace)
               if result_name != '':
                  sh.locals[result_name].append(result)
               progress.step()
         finally:
            progress.kill()
      if result_name != '':
         sh.run(result_name)

   def _run_in_shell(self, sh, result_name, func_call):
      if len(self.images) == 1:
         if result_name != '':
            source = '%s = ' % result_name
         else:
            source = ''
         source += '.'.join((self.image_name, func_call))
         sh.run(source)
      else:
         if result_name != '':
            sh.run('%s = []' % result_name)
         sh.run('for i in range(len(%s)):' % self.image_name)
         source = '%s[i].%s' % (self.image_name, func_call)
         if result_name != '':
            source = '%s.append(%s)' % (result_name, source)
         sh.run('\t' + source)
         sh.run('\n')



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Mixin classes to add wrapper-generation support for arguments list

import re
from enums import *
import util

class Arg:
   arg_format = 'O'
   convert_from_PyObject = False
   multiple = False
   delete_cpp = False
   uid = 0

   def __getitem__(self, attr):
      return getattr(self, attr)

   def __iter__(self):
      raise RuntimeError()

   def _get_symbol(self):
      return re.sub(r"\s", "_", self.name) + '_arg'
   symbol = property(_get_symbol)

   def _get_pysymbol(self):
      if self.convert_from_PyObject:
         return re.sub(r"\s", "_", self.name) + '_pyarg'
      else:
         return self.symbol
   pysymbol = property(_get_pysymbol)

   def declare(self):
      if self.name == None:
         self.name = "_%08d" % Arg.uid
         Arg.uid += 1
      if self.name == 'return' and hasattr(self, 'return_type'):
         result = "%s %s;\n" % (self.return_type, self.symbol)
      else:
         result = "%s %s;\n" % (self.c_type, self.symbol)
      if self.convert_from_PyObject:
         result += "PyObject* %s;\n" % self.pysymbol
      return result

   def from_python(self):
      return ""

   def _do_call(self, function, output_args):
      if function.progress_bar:
         output_args.append('ProgressBar((char *)"%s")' % function.progress_bar);
      if function.feature_function:
         return "%s(%s, feature_buffer);" % (function.__name__, ", ".join(output_args))
      else:
         if function.return_type != None:
            lhs = function.return_type.symbol + " = "
         else:
            lhs = ""
         rhs = "%s(%s)" % (function.__name__, ", ".join(output_args))
         if function.return_type.__class__.__name__ == "Pixel":
            rhs = "pixel_to_python(%s)" % rhs
         return "%s%s;\n" % (lhs, rhs)

   def call(self, function, args, output_args, limit_choices=None):
      if len(args):
         return args[0].call(function, args[1:], output_args + [self.symbol], limit_choices)
      else:
         return self._do_call(function, output_args + [self.symbol])

   def delete(self):
      if self.delete_cpp:
         return "delete %(symbol)s;" % self
      else:
         return ""

class Int(Arg):
   arg_format = 'i'
   c_type = 'int'

   def to_python(self):
      return '%(pysymbol)s = PyInt_FromLong((long)%(symbol)s);' % self

Choice = Check = Int

class Float(Arg):
   arg_format = 'd'
   c_type = 'double'

   def to_python(self):
      return '%(pysymbol)s = PyFloat_FromDouble((double)%(symbol)s);' % self

class Complex(Arg):
   arg_format = 'D'
   c_type = 'Py_complex'

   def to_python(self):
      return '''%(pysymbol)s = PyComplex_FromDoubles(%(symbol)s.real(), %(symbol)s.imag());''' % self

   def from_python(self):
      return '''Py_complex %(pysymbol)s_temp = PyComplex_AsCComplex(%(pysymbol)s);
      %(symbol)s = ComplexPixel(%(pysymbol)s_temp.real, %(pysymbol)s_temp.imag);'''

class String(Arg):
   arg_format = 's'
   c_type = 'char*'
   return_type = 'std::string'

   def to_python(self):
      return "%(pysymbol)s = PyString_FromStringAndSize(%(symbol)s.data(), %(symbol)s.size());" % self

FileOpen = FileSave = Directory = ChoiceString = String

class ImageType(Arg):
   c_type = 'Image*'
   convert_from_PyObject = True
   multiple = True

   def from_python(self):
      return """if (!is_ImageObject(%(pysymbol)s)) {
          PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be an image");
          return 0;
        }
        %(symbol)s = ((Image*)((RectObject*)%(pysymbol)s)->m_x);
        image_get_fv(%(pysymbol)s, &%(symbol)s->features, &%(symbol)s->features_len);
        """ % self

   def to_python(self):
      return "%(pysymbol)s = create_ImageObject(%(symbol)s);" % self

   def call(self, function, args, output_args, limit_choices=None):
      if function.image_types_must_match and limit_choices is not None:
         choices = self._get_choices_for_pixel_type(limit_choices)
      else:
         choices = self._get_choices()
      result = "switch(get_image_combination(%(pysymbol)s)) {\n" % self
      for choice, pixel_type in choices:
         result += "case %s:\n" % choice.upper()
         new_output_args = output_args + ["*((%s*)%s)" % (choice, self.symbol)]
         if len(args) == 0:
            result += self._do_call(function, new_output_args)
         else:
            if limit_choices is None:
               result += args[0].call(function, args[1:], new_output_args, pixel_type)
            else:
               result += args[0].call(function, args[1:], new_output_args, limit_choices)
         result += "break;\n"
      result += "default:\n"
      acceptable_types = [util.get_pixel_type_name(y).upper() for x, y in choices]
      if len(acceptable_types) >= 2:
         phrase = "values are"
         acceptable_types[-1] = "and " + acceptable_types[-1]
      else:
         phrase = "value is"
      acceptable_types = ", ".join(acceptable_types)
      result += ('PyErr_Format(PyExc_TypeError,'
                 '"The \'%s\' argument of \'%s\' can not have pixel type \'%%s\'. '
                 'Acceptable %s %s."'
                 ', get_pixel_type_name(%s));\nreturn 0;\n' %
                 (self.name, function.__name__, phrase, acceptable_types, self.pysymbol))
      result += "}\n"
      return result

   def _get_choices(self):
      result = []
      pixel_types = list(self.pixel_types[:])
      pixel_types.sort()
      for type in pixel_types:
         result.extend(self._get_choices_for_pixel_type(type))
      return result

   def _get_choices_for_pixel_type(self, pixel_type):
      if pixel_type == ONEBIT:
         result = ["OneBitImageView", "Cc", "OneBitRleImageView", "RleCc", "MlCc"]
      else:
         result = [util.get_pixel_type_name(pixel_type) + "ImageView"]
      return [(x, pixel_type) for x in result]

class Rect(Arg):
   c_type = 'Rect*'
   convert_from_PyObject = True

   def from_python(self):
      return """
      if (!is_RectObject(%(pysymbol)s)) {
        PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a Rect");
        return 0;
      }
      %(symbol)s = (((RectObject*)%(pysymbol)s)->m_x);
      """ % self

   def to_python(self):
      return """
      return_pyarg = create_RectObject(*%(symbol)s);
      delete %(symbol)s;""" % self
      #return "%(pysymbol)s = create_RectObject(%(symbol)s);" % self

class Region(Arg):
   c_type = 'Region*'
   convert_from_PyObject = True

   def from_python(self):
      return """
      if (!is_RegionObject(%(pysymbol)s)) {
        PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a Region");
        return 0;
      }
      %(symbol)s = (Region*)((RectObect*)%(pysymbol)s)->m_x;""" % self

   def to_python(self):
      return """
      return_pyarg = create_RegionObject(*%(symbol)s);
      delete %(symbol)s;""" % self

class RegionMap(Arg):
   c_type = 'RegionMap*'
   convert_from_PyObject = True

   def from_python(self):
      return """
      if (!is_RegionMapObject(%(pysymbol)s)) {
        PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a RegionMap.");
        return 0;
      }
      %(symbol)s = (RegionMap*)((RegionMapObect*)%(pysymbol)s)->m_x;""" % self

   def to_python(self):
      return """
      return_pyarg = create_RegionMapObject(*%(symbol)s);
      delete %(symbol)s;""" % self

class ImageList(Arg):
   c_type = 'ImageVector'
   return_type = 'std::list<Image*>*'
   convert_from_PyObject = True

   def from_python(self):
      return """
          const char* type_error_%(name)s = "Argument '%(name)s' must be an iterable of images.";
          PyObject* %(pysymbol)s_seq = PySequence_Fast(%(pysymbol)s, type_error_%(name)s);
          if (%(pysymbol)s_seq == NULL)
            return 0;
          int %(symbol)s_size = PySequence_Fast_GET_SIZE(%(pysymbol)s_seq);
          %(symbol)s.resize(%(symbol)s_size);
          for (int i=0; i < %(symbol)s_size; ++i) {
            PyObject *element = PySequence_Fast_GET_ITEM(%(pysymbol)s_seq, i);
            if (!is_ImageObject(element)) {
              PyErr_SetString(PyExc_TypeError, type_error_%(name)s);
              return 0;
            }
            %(symbol)s[i] = std::pair<Image*, int>((Image*)(((RectObject*)element)->m_x), get_image_combination(element));
            image_get_fv(element, &%(symbol)s[i].first->features,
                         &%(symbol)s[i].first->features_len);
          }
          Py_DECREF(%(pysymbol)s_seq);""" % self

   def to_python(self):
      return """
      %(pysymbol)s = ImageList_to_python(%(symbol)s);
      delete %(symbol)s;
      """ % self

class Class(Arg):
   arg_format = 'O'
   c_type = 'PyObject*'
   convert_from_PyObject = True

   def from_python(self):
      return "%(symbol)s = %(pysymbol)s;" % self

   def to_python(self):
      return "%(pysymbol)s = %(symbol)s;" % self

class IntVector(Arg):
   arg_format = 'O'
   convert_from_PyObject = True
   c_type = 'IntVector*'
   delete_cpp = True

   def from_python(self):
      return """
      %(symbol)s = IntVector_from_python(%(pysymbol)s);
      if (%(symbol)s == NULL) return 0;
      """ % self

   def to_python(self):
      return """
      %(pysymbol)s = IntVector_to_python(%(symbol)s);
      delete %(symbol)s;
      """ % self

class FloatVector(Arg):
   arg_format = 'O'
   convert_from_PyObject = True
   c_type = 'FloatVector*'
   delete_cpp = True

   def from_python(self):
      return """
      %(symbol)s = FloatVector_from_python(%(pysymbol)s);
      if (%(symbol)s == NULL) return 0;
      """ % self

   def to_python(self):
      return """
      %(pysymbol)s = FloatVector_to_python(%(symbol)s);
      delete %(symbol)s;
      """ % self

class ComplexVector(Arg):
   arg_format = 'O'
   convert_from_PyObject = True
   c_type = 'ComplexVector*'
   delete_cpp = True

   def from_python(self):
      return """
      %(symbol)s = ComplexVector_from_python(%(pysymbol)s);
      if (%(symbol)s == NULL) return 0;
      """ % self

   def to_python(self):
      return """
      %(pysymbol)s = ComplexVector_to_python(%(symbol)s);
      delete %(symbol)s;
      """ % self

class Pixel(Arg):
   def from_python(self):
      return ''

   def call(self, function, args, output_args, limit_choices=None):
      if limit_choices is None:
         raise RuntimeError("You can not create a plugin that takes a Pixel argument that does not have a self type")
      pixel_type = limit_choices
      new_output_args = (output_args +
                         ["pixel_from_python<%sPixel>::convert(%s)" %
                          (util.get_pixel_type_name(pixel_type), self.pysymbol)])
      if len(args) == 0:
         return self._do_call(function, new_output_args)
      else:
         return args[0].call(function, args[1:], new_output_args, limit_choices)

class Point(Arg):
   c_type = "Point"
   delete_cpp = False
   convert_from_PyObject = True

   def from_python(self):
      return """
      try {
         %(symbol)s = coerce_Point(%(pysymbol)s);
      } catch (std::invalid_argument e) {
         PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a Point, or convertible to a Point");
         return 0;
      }
      """ % self

   def to_python(self):
      return """%(pysymbol)s = create_PointObject(%(symbol)s);""" % self

class FloatPoint(Arg):
   c_type = "FloatPoint"
   delete_cpp = False
   convert_from_PyObject = True

   def from_python(self):
      return """
      try {
         %(symbol)s = coerce_FloatPoint(%(pysymbol)s);
      } catch (std::invalid_argument e) {
         PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a FloatPoint object, or convertible to a FloatPoint object");
         return 0;
      }
      """ % self

   def to_python(self):
      return """%(pysymbol)s = create_FloatPointObject(%(symbol)s);""" % self

class Dim(Arg):
   c_type = 'Dim'
   convert_from_PyObject = True

   def from_python(self):
      return """
      if (!is_DimObject(%(pysymbol)s)) {
        PyErr_SetString(PyExc_TypeError, "Argument '%(name)s' must be a Dim object");
        return 0;
      }
      %(symbol)s = *((Dim*)((DimObject*)%(pysymbol)s)->m_x);""" % self

   def to_python(self):
      return """
      return_pyarg = create_DimObject(*%(symbol)s);
      delete %(symbol)s;""" % self

class PointVector(Arg):
   arg_format = 'O'
   convert_from_PyObject = True
   c_type = 'PointVector*'
   delete_cpp = True

   def from_python(self):
      return """
      %(symbol)s = PointVector_from_python(%(pysymbol)s);
      if (%(symbol)s == NULL) return 0;
      """ % self

   def to_python(self):
      return """
      %(pysymbol)s = PointVector_to_python(%(symbol)s);
      delete %(symbol)s;
      """ % self

class ImageInfo(Arg):
   arg_format = "O";
   c_type = 'ImageInfo*'
   convert_from_PyObject = True

   def to_python(self):
      return "%(pysymbol)s = create_ImageInfoObject(%(symbol)s);""" % self

from gamera import args
args.mixin(locals(), "C++ Wrappers")



#
# Copyright (C) 2009-2015 Christoph Dalitz
#               2010      Oliver Christen, Tobias Bolten
#               2011      Christian Brandt
#               2015      Manuel Jeltsch
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
from gamera.args import NoneDefault
import _geometry
try:
  from gamera.core import RGBPixel
except Exception:
  def RGBPixel(*args):
    pass


class voronoi_from_labeled_image(PluginFunction):
  u"""
  Computes the area Voronoi tesselation from an image containing labeled
  Cc's. In the returned onebit image, every pixel is labeled with the
  label value of the closest Cc in the input image.

  To prepare the input image, you can use cc_analysis__. When the Cc's
  only consist of single points, the area Voronoi tesselation is identical
  to the ordinary Voronoi tesselation.

  .. __: segmentation.html#cc-analysis

  The implementation applies a watershed algorithm to the distance transform
  of the input image, a method known as *seeded region growing* (U. K\u00f6the:
  *Primary Image Segmentation.* Proceedings 17th DAGM-Symposium, pp. 554-561,
  Springer, 1995).

  The example shown below is the image *voronoi_cells* as created with
  the following code:

  .. code:: Python

    # create an area Voronoi tesselation and
    # mark the cells and their edges in color
    ccs = image.cc_analysis()  # labels the image
    voronoi = image.voronoi_from_labeled_image()
    voronoi_cells = voronoi.color_ccs()
    voronoi_cells.highlight(image, RGBPixel(0,0,0))
    voronoi_edges = voronoi.labeled_region_edges()
    voronoi_cells.highlight(voronoi_edges, RGBPixel(255,255,255))
  """
  self_type = ImageType([ONEBIT,GREYSCALE])
  return_type = ImageType([ONEBIT,GREYSCALE])
  def __doc_example1__(images):
    image = images[ONEBIT]
    ccs = image.cc_analysis()
    voronoi = image.voronoi_from_labeled_image()
    voronoi_cells = voronoi.color_ccs()
    voronoi_cells.highlight(image, RGBPixel(0,0,0))
    voronoi_edges = voronoi.labeled_region_edges()
    voronoi_cells.highlight(voronoi_edges, RGBPixel(255,255,255))
    return [image, voronoi_cells]
  doc_examples = [__doc_example1__]
  author = u"Christoph Dalitz, based on code by Ullrich K\u00f6the"

class voronoi_from_points(PluginFunction):
  """
  Computes the Voronoi tesselation from a list of points and point labels.
  The result is directly written to the input image. Each white pixel is
  labeled with the label value of the closest point. Non white pixel in the
  input image are not overwritten.

  The arguments *points* and *labels* specify the points and labels, such
  that ``labels[i]`` is the label of ``points[i]``. Note that the labels
  need not necessarily all be different, which can be useful as an 
  approximation of an area Voronoi tesselation.

  The algorithm is very simple: it stores the points in a `kd-tree`_ and
  then looks up the closest point for each image pixel. This has a runtime
  of *O(N log(n))*, where *N* is the number of image pixels and *n* is the
  number of points. For not too many points, this should be faster than the
  morphological region growing approach of `voronoi_from_labeled_image`_.

.. _`kd-tree`: kdtree.html
.. _`voronoi_from_labeled_image`: #voronoi-from-labeled-image

  The example shown below is the image *voronoi_edges* as created with
  the following code:

  .. code:: Python

    # create a Voronoi tesselation and mark
    # the cell edges in a second image
    points = [(10,10),(20,30),(32,22),(85,14),(40,70),(80,85)]
    voronoi = Image((0,0),(90,90))
    voronoi.voronoi_from_points(points,[i+2 for i in range(len(points))])
    voronoi_edges = voronoi.labeled_region_edges()
    for p in points:
       voronoi_edges.set(p,1)
  """
  self_type = ImageType([ONEBIT,GREYSCALE])
  args = Args([PointVector("points"),IntVector("labels")])
  return_type = None
  def __doc_example1__(images):
    from gamera.core import Image
    points = [(10,10),(20,30),(32,22),(85,14),(40,70),(80,85)]
    voronoi = Image((0,0),(90,90))
    voronoi.voronoi_from_points(points,[i+2 for i in range(len(points))])
    voronoi_edges = voronoi.labeled_region_edges()
    for p in points:
      voronoi_edges.set(p,1)
    return [voronoi_edges]
  doc_examples = [__doc_example1__]
  author = "Christoph Dalitz"

class labeled_region_neighbors(PluginFunction):
  """
  For an image containing labeled regions, a list of all label pairs belonging
  to touching regions is returned. When *eight_connectivity* is ``True``
  (default), 8-connectivity is used for neighborship, otherwise
  4-connectivity is used.

  This can be useful for building a Delaunay graph from a Voronoi tesselation
  as in the following example:

  .. code:: Python

    #
    # build Delaunay graph of neighboring (i.e. adjacent) Cc's
    #

    # create map label->Cc for faster lookup later
    ccs = image.cc_analysis()
    label_to_cc = {}
    for c in ccs:
       label_to_cc[c.label] = c

    # compute area Voronoi tesselation and neighborship list
    voronoi = image.voronoi_from_labeled_image()
    labelpairs = voronoi.labeled_region_neighbors()

    # build map of all neighbors for each label for fast lookup
    neighbors = {}
    for label in label_to_cc.keys():
       neighbors[label] = []
    for n in labelpairs:
       neighbors[n[0]].append(n[1])
       neighbors[n[1]].append(n[0])

    # now, all neighbors to a given cc can be looked up with
    neighbors_of_cc = [label_to_cc[label] for label in neighbors[cc.label]]

  """
  self_type = ImageType([ONEBIT])
  args = Args([Check("eight_connectivity",default=True)])
  return_type = Class("labelpairs")
  author = "Christoph Dalitz"
  # wrapper for passing default argument
  def __call__(self, eight_connectivity=True):
      return _geometry.labeled_region_neighbors(self, eight_connectivity)
  __call__ = staticmethod(__call__)

class delaunay_from_points(PluginFunction):
  """
  Computes the Delaunay triangulation directly from a list of points and
  point labels. The result is a list which contains tuples of adjacent labels,
  where in each tuple the smaller label is given first.

  The arguments *points* and *labels* specify the points and nonnegative
  labels, such that ``labels[i]`` is the label of ``points[i]``. Note that
  the labels need not necessarily all be different, which can be useful
  for the computation of a neighborship graph from a set of connected
  components.

  The computation of the Delaunay triangulation is based on the Delaunay
  tree which is a randomized geometric data structure. It is
  described in O. Devillers, S. Meiser, M. Teillaud:
  `Fully dynamic Delaunay triangulation in logarithmic expected time per operation.`__
  Computational Geometry: Theory and Applications 2, pp. 55-80, 1992.

.. __: http://hal.inria.fr/inria-00090678

  This can be useful for building a neighborhood graph as shown in the
  following example:

  .. code:: Python

    from gamera import graph
    from gamera.plugins.geometry import delaunay_from_points
    
    points = [(10,10),(20,30),(32,22),(85,14),(40,70),(80,85)]
    labels = range(len(points))
    neighbors = delaunay_from_points(points, labels)
    
    g = graph.Graph(graph.UNDIRECTED)
    for pair in neighbors:
        g.add_edge(pair[0], pair[1])
  """
  self_type = None
  args = Args([PointVector("points"), IntVector("labels")])
  return_type = Class("labelpairs")
  author = "Oliver Christen (based on code by Olivier Devillers)"


class graph_color_ccs(PluginFunction):
    """
    Returns an RGB Image where each segment is colored with one of the colors
    from *colors* in such a way that neighboring segments have different
    colors. Optionally, each segment can also be colored uniquely with
    a color that is close to one of the given *colors*.

    Reference:

      C. Dalitz, T. Bolten, O. Christion:
      `\"Color Visualization of 2D Segmentations.\"`__
      International Conference on Information Visualization Theory
      and Applications (IVAPP), pp. 567-572 (2013)

.. __: http://informatik.hsnr.de/~dalitz/data/publications/ivapp13-coloring.pdf

    This function can be used to verify that the pagesegmentation 
    e.g. ``runlength_smearing`` is working correctly for your image.


    Arguments:
    
    *ccs*:
        ImageList which contains ccs to be colored. Must be views on
        the image on which this method is called.

    *colors*:
        list of colors (instances of RGBPixel) which will be used for coloring.
        When ``None``, the default set of seven colors given in the example
        below is used.

    *method*:
        Controls the calculation of the neighborhood graph:

            0 = from the CC center points
            (fastest, but can be inaccurate for large CC's)

            1 = from a 20 percent sample of the contour points
            (reasonable compromise between speed and accuracy)

            2 = from the exact area Voronoi diagram
            (can be slow on large images)

    *unique*:
        When ``True``, each segment obtains a unique color that is
        close to one of the colors in *colors* with the restriction
        that neighboring segments obtain sufficiently different colors.
        See the reference above for details.

    Example:

    .. code:: Python

       ccs = imgage.cc_analysis()
       colors = [ RGBPixel(150, 0, 0),
                  RGBPixel(0, 250, 0),
                  RGBPixel(0, 0, 255),
                  RGBPixel(250, 0, 255),
                  RGBPixel(50, 150, 50),
                  RGBPixel(0, 190, 255),
                  RGBPixel(230, 190, 20) ]
       rgb = imgage.graph_color_ccs(ccs, colors, 1)

    .. note:: *colors* must not contain less than six colors.

    """
    category = "Color"
    author = "Oliver Christen, Tobias Bolten, Christoph Dalitz"
    args = Args([ImageList('ccs'), Class('colors', klass=RGBPixel, list_of=True,default=NoneDefault), Choice('method', ["CC center", "20% contour points", "voronoi diagram"], default=1), Check("unique", default=False)])
    self_type = ImageType([ONEBIT])
    return_type = ImageType([RGB])

    def __call__(image, ccs, colors=None, method=1, unique=False):
      if colors == None:
        from gamera.core import RGBPixel
        colors = [ RGBPixel(150, 0, 0),
                   RGBPixel(0, 250, 0),
                   RGBPixel(0, 0, 255),
                   RGBPixel(250, 0, 255),
                   RGBPixel(50, 150, 50),
                   #RGBPixel(120, 120, 120),
                   #RGBPixel(250, 250, 0),
                   RGBPixel(0, 190, 255),
                   RGBPixel(230, 190, 20),
                   ]
      return _geometry.graph_color_ccs(image, ccs, colors, method, unique)
    __call__ = staticmethod(__call__)


class convex_hull_from_points(PluginFunction):
    """Returns the polygon vertices of the convex hull of the given list of
points.

The function uses Graham's scan algorithm as described e.g. in Cormen et al.:
*Introduction to Algorithms.* 2nd ed., MIT Press, p. 949, 2001
"""
    self_type = None
    args = Args([PointVector("points")])
    return_type = PointVector("convexhull")
    author = "Christian Brandt, Christoph Dalitz, and David Kolanus"


class convex_hull_as_points(PluginFunction):
   """Returns the vertex points of the convex hull of all black pixels
in the given image.

Actually not all black pixels are required for computing the convex hull,
but only the left and right contour pixels of the image. This follows
from the fact that, when a point is invisible both from the left and the
right, it lies on the connection axis between two visible points and thus
cannot be a vertex point of the convex hull.
   """
   self_type = ImageType([ONEBIT])
   args = Args([])
   return_type = PointVector("convexhull")
   author = "Christoph Dalitz"


class convex_hull_as_image(PluginFunction):
   """Returns an image containing the polygon of the convex hull calculated
by convex_hull_as_points_.

.. _convex_hull_as_points: geometry.html#convex_hull_as_points
   """
   self_type = ImageType([ONEBIT])
   args = Args([Check("filled",default=False)])
   return_type = ImageType([ONEBIT])
   author = "Christoph Dalitz"
   doc_examples = [(ONEBIT,)]

   def __call__(image, filled=False):
       return _geometry.convex_hull_as_image(image, filled)
   __call__ = staticmethod(__call__)


class max_empty_rect(PluginFunction):
   """Returns the maximum area empty rect that fits into the image without
containing any of the black image pixels. This problem is in the literature
generally known as the *Largest Empty Rectangle Problem*.

When the image does not contain a white pixel at all, an exception of type
``RuntimeError`` is thrown. The coordinates of the returned rectangle are
relative to the upper left corner of the image.

Reference: D. Vandevoorde: `\"The Maximal Rectangle Problem.\"`__ Dr. Dobb's,
April 1998.

.. __: http://www.drdobbs.com/database/184410529
   """
   self_type = ImageType([ONEBIT])
   args = Args([])
   return_type = Rect("max_epmty_rect")
   author = "Daveed Vandevoorde and Christoph Dalitz"
   def __doc_example1__(images):
       from gamera.core import Image
       from gamera.core import Point as P
       img = Image((0,0),(90,90))
       points = [P(10,10),P(20,30),P(32,22),P(85,14),P(40,70),P(80,85)]
       for p in points:
           img.draw_filled_rect((p.x-2,p.y-2),(p.x+1,p.y+1),1)
       r = img.max_empty_rect()
       rgb = img.to_rgb()
       rgb.draw_hollow_rect(r,RGBPixel(255,0,0))
       rgb.highlight(img,RGBPixel(0,0,0))
       return [rgb]
   doc_examples = [__doc_example1__]

class hough_lines(PluginFunction):
    """
    Finds lines in a point set with the Hough transform. Lines are 
    represented by the Hessian normal form, i.e. by the two parameters
    *theta* and *rho*, where *rho* is the distance of the line to the origin,
    and *theta* is the angle of normal; a horizontal line therefore
    has *theta* = 90 degrees.

    The found lines are returned as a list of tuples 
    [(*votes*, *theta*, *rho*),...],
    where *votes* is the number of points from the point set belonging to
    this line. The lines are sorted by *votes* in descending order. Which and
    how many lines are returned is controlled with *threshold* and *n_lines*.
    The angles *theta* are given in degrees.
    
    Arguments:
    
    *points*:
        The points from the point cloud.

    *theta_min*, *theta_step*, *theta_max*
        Theta-quantization of the Hough space, i.e. the range of angles
        to be tested. Angles are spezified in degrees.

    *rho_min*, *rho_step*, *rho_max*
        Rho-quantization of the Hough space, i.e. the range of origin-distances
        to be tested.

    *n_lines*
        The number of highest votes lines that are returned. When set to
        zero, all found lines are returned.

    *threshold*
        Only lines with more votes (points) than *threshold* are returned.

    """
    author = "Manuel Jeltsch and Christoph Dalitz"
    return_type = Class("votes_theta_rho")
    self_type = None
    args = Args([PointVector("points"), 
                Real("theta_min"), Real("theta_step"), Real("theta_max"),
                Real("rho_min"), Real("rho_step"), Real("rho_max"),
                Int("n_lines"), Real("threshold")])


class GeometryModule(PluginModule):
  cpp_headers = ["geometry.hpp"]
  category = "Geometry"
  import glob
  cpp_sources=["src/geostructs/kdtree.cpp", "src/geostructs/delaunaytree.cpp"] + glob.glob("src/graph/*.cpp")
  functions = [voronoi_from_labeled_image,
               voronoi_from_points,
               labeled_region_neighbors,
               delaunay_from_points,
               graph_color_ccs,
               convex_hull_from_points,
               convex_hull_as_points,
               convex_hull_as_image,
               max_empty_rect,
               hough_lines]
  author = "Christoph Dalitz"
  url = "http://gamera.sourceforge.net/"

module = GeometryModule()

delaunay_from_points = delaunay_from_points()
convex_hull_from_points = convex_hull_from_points()
hough_lines = hough_lines()



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2002-2003 Karl MacMillan and Michael Droettboom
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera import core, knn, graph, gamera_xml
core.init_gamera()

def get_lengths(node, depth, lengths, cur_depth=0, path = {}):
   if cur_depth >= depth:
      return
   for edge in node.edges:
      if path.has_key(edge):
         continue
      path[edge] = None
      lengths.append(edge.cost)
      get_lengths(edge.traverse(node), depth, lengths, cur_depth + 1, path)

def label(graph, node, label_start, label):
   for node in graph.DFS(node):
      node().classify_automatic(label_start + str(label))
                        
def make_subtrees_stddev(graph, ratio, distance, relabel=1, lab="cluster."):
   import stats
   cur_label = 0
   remove = []
   i = 0
   for edge in graph.get_edges():
      lengths = []
      path = {}
      #print node().get_main_id(), edge.cost
      get_lengths(edge.from_node, distance, lengths, 0, path)
      lengths.remove(edge.cost)
      #print lengths
      if not (len(lengths) > 1):
         continue
      mean = stats.mean(lengths)
      stdev2 = stats.samplestdev([mean, edge.cost])
      #print mean, stdev2, edge.cost, len(lengths)
      if stdev2 > ratio:
         #graph.remove_edge(edge)
         remove.append(edge)

   for edge in remove:
      graph.remove_edge(edge)

   if relabel:
      cur_label = 0
      for node in graph.get_nodes():
         node().classify_manual("")
      for node in graph.get_nodes():
         if node().get_main_id() == "":
            label(graph, node, lab, cur_label)
            cur_label += 1
   nodes = []

   for node in graph.get_nodes():
      nodes.append(node())

   return nodes

def make_spanning_tree(glyphs, k=None):
   if k is None:
      k = knn.kNNInteractive()
   uniq_dists = k.distance_matrix(glyphs, 0)
   g = graph.Undirected()
   g.create_minimum_spanning_tree(glyphs, uniq_dists)
   return g

def cluster(glyphs, ratio=1.0, distance=2, label="cluster.", k=None, relabel=1):
   g = make_spanning_tree(glyphs, k)

   return make_subtrees_stddev(g, ratio, distance, lab=label)


def cluster2(glyphs):
   ko = knn.kNNInterative()
   gc = knn.glyphs_by_category(cluster(glyphs, k=ko))
   small = []
   large = []
   for x in gc.itervalues():
      if len(x) < 10:
         small.extend(x)
      else:
         large.append(x)
   print len(small)
   output = cluster(small, 1, 1, label="cluster_small.", k=ko)
   cur_label = 0
   for x in large:
      l = x[0].get_main_id() + ".cluster_large" + str(cur_label) + "."
      cur_label += 1
      c = cluster(x, .6, 4, label=l, k=ko)
      output.extend(c)
   return output

def do_tests(filename):
   from gamera import gamera_xml
   from gamera import core
   import os.path

   image = core.load_image(filename)
   glyphs = image.cc_analysis()
   g = make_spanning_tree(glyphs)
   g_orig = g.copy()
   c = make_subtrees_stddev(g, 1.5)
   gamera_xml.glyphs_to_xml(c, os.path.abspath(os.path.dirname(filename) + "cluster_1_5_" + os.path.basename(filename)))
   g = g_orig.copy()
   c = make_subtrees_stddev(g, 2.0)
   gamera_xml.glyphs_to_xml(c, os.path.abspath(os.path.dirname(filename) + "cluster_2_0_" + os.path.basename(filename)))
   g = g_orig.copy()
   c = make_subtrees_stddev(g, 2.5)
   gamera_xml.glyphs_to_xml(c, os.path.abspath(os.path.dirname(filename) + "cluster_2_5_" + os.path.basename(filename)))
   g = g_orig.copy()
   c = make_subtrees_stddev(g, 3.0)
   gamera_xml.glyphs_to_xml(c, os.path.abspath(os.path.dirname(filename) + "cluster_3_0_" + os.path.basename(filename)))

def make_unique_names(glyphs):
   for i in range(len(glyphs)):
      glyphs[i].classify_automatic(glyphs[i].get_main_id() + str(i))

def graphvis_output(G, filename):
   fd = open(filename, 'w')
   if G.is_directed():
      fd.write("digraph G {\n")
      for node in G.get_nodes():
         for edge in node.out_edges:
            fd.write('   %s -> %s [ label = %f ];\n' % (node(), edge.to_node(), edge()))
      fd.write("}\n")
   else:
      fd.write("graph G {\n")
      for edge in G.get_edges():
         fd.write('   "%s" -- "%s" [label="%.2f",len="%f"];\n'
                  % (edge.from_node().get_main_id(), edge.to_node().get_main_id(),
                     edge.cost, edge.cost / 4))
      fd.write("}\n")
   fd.close()

# this function is a stub and does not do anything
# unless we figure out what it was supposed to do,
# I have commented it out (CD)
# def analysis(glyphs):
#    by_id = {}
#    for x in glyphs:
#       id = x.get_main_id()
#       if not by_id.has_key(id):
#          by_id[id] = []
#       by_id[id].append(x)
#    num_features = len(x.features)

#    for x in by_id:
#       for i in range(len(x)):
#          sum_vec
   



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

def glyphs_comma_delim(glyphs, filename):
    """Export the id_name and features of a list of
    glyphs to a comma delimited format"""
    file = open(filename, "w")
    for x in glyphs:
        file.write(x.id_name[0][1])
        for f in x.features:
            file.write(',')
            file.write(str(f))
        file.write('\n')
    file.close()

def glyphs_space_delim(glyphs, filename):
    """Export the id_name and features of a list of
    glyphs to a space delimited format"""
    file = open(filename, "w")
    for x in glyphs:
        file.write(x.id_name[0][1])
        for f in x.features:
            file.write(' ')
            file.write(str(f))
        file.write('\n')
    file.close()



from gamera.core import *
from math import sqrt

def test_rect():
   r1 = Rect(Point(25, 25), Point(74, 64))
   assert r1.ul == Point(25, 25)
   assert r1.ul_x == 25
   assert r1.ul_y == 25
   assert r1.lr == Point(74, 64)
   assert r1.lr_x == 74
   assert r1.lr_y == 64
   assert r1.size == Size(49, 39)
   assert r1.dim == Dim(50, 40)
   assert r1.nrows == 40
   assert r1.ncols == 50
   assert r1.height == 39
   assert r1.width == 49
   assert r1.ur == Point(74, 25)
   assert r1.ur_x == 74
   assert r1.ur_y == 25
   assert r1.ll == Point(25, 64)
   assert r1.ll_x == 25
   assert r1.ll_y == 64

   assert r1.center == Point(49, 44)
   assert r1.center_x == 49
   assert r1.center_y == 44

   assert r1.contains_x(40)
   assert r1.contains_y(40)
   assert not r1.contains_x(24)
   assert not r1.contains_y(24)
   assert r1.contains_point(Point(74, 64))
   assert not r1.contains_point(Point(75, 65))

   assert Rect(Point(0,0), Point(100,100)).contains_rect(r1)
   assert r1.contains_rect(Rect(Point(30,30), Point(30,40)))
   assert not r1.contains_rect(Rect(Point(20,30), Point(60,40)))

   assert Rect(Point(0,0), Point(100,100)).intersects(r1)
   assert r1.intersects(Rect(Point(30,30), Point(60,70)))
   assert r1.intersects(Rect(Point(20,30), Point(60,40)))
   assert not r1.intersects(Rect(Point(0,0), Point(20,20)))

   assert r1.union_rects([r1, Rect(Point(0,0), Point(20,20))]) == Rect(Point(0,0), Point(74,64))

   distance_euclid = r1.distance_euclid(Rect(Point(0,0), Point(25,25)))
   assert distance_euclid > 48.9 and distance_euclid < 49.0

   assert r1.distance_bb(Rect(Point(0,0), Point(20,20))) == sqrt(5*5 + 5*5)
   assert r1.distance_cx(Rect(Point(0,0), Point(20,20))) == 39
   assert r1.distance_cy(Rect(Point(0,0), Point(20,20))) == 34



#
# Copyright (C) 2005      John Ashley Burgoyne and Ichiro Fujinaga
#               2012      Andrew Hankinson
#               2011-2012 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# TODO: Add GREY16 compatibility.
# TODO: Add Yanowitz and Bruckstein post-processing (a la Trier and Jain).

"""Adaptive binarization tools."""

from gamera.plugin import *
from gamera.args import NoneDefault
import _binarization

class image_mean(PluginFunction):
    """
    Returns the mean over all pixels of an image as a FLOAT.
    """
    category = "Binarization/RegionInformation"
    return_type = Real("output")
    self_type = ImageType([GREYSCALE,GREY16,FLOAT])
    def __call__(self):
        return _binarization.image_mean(self)
    __call__ = staticmethod(__call__)


class image_variance(PluginFunction):
    """
    Returns the variance over all pixels of an image as a FLOAT.
    """
    category = "Binarization/RegionInformation"
    return_type = Real("output")
    self_type = ImageType([GREYSCALE,GREY16,FLOAT])
    def __call__(self):
        return _binarization.image_variance(self)
    __call__ = staticmethod(__call__)


class mean_filter(PluginFunction):
    """
    Returns the regional mean of an image as a FLOAT.

    *region_size*
      The size of the region in which to calculate a mean.
    """
    category = "Binarization/RegionInformation"
    return_type = ImageType([FLOAT], "output")
    self_type = ImageType([GREYSCALE,GREY16,FLOAT])
    args = Args([Int("region size", default=5)])
    doc_examples = [(GREYSCALE,), (GREY16,), (FLOAT,)]
    def __call__(self, region_size=5):
        return _binarization.mean_filter(self, region_size)
    __call__ = staticmethod(__call__)


class variance_filter(PluginFunction):
    """
    Returns the regional variance of an image as a FLOAT.

    *means*
      Pre-calculated means for each region.

    *region_size*
      The size of the region in which to calculate the variance.
    """
    category = "Binarization/RegionInformation"
    return_type = ImageType([FLOAT], "output")
    self_type = ImageType([GREYSCALE,GREY16,FLOAT])
    args = Args([ImageType([FLOAT], "means"),
                 Int("region size", default=5)])
    def __call__(self, means, region_size=5):
        return _binarization.variance_filter(self, means, region_size)
    __call__ = staticmethod(__call__)


class wiener_filter(PluginFunction):
    """
    Adaptive Wiener filter for de-noising.

    See:
    
    J. Lim. 2001. *Two-Dimensional Signal Processing.* Englewood
    Cliffs: Prentice Hall.

    *region_size*
      The size of the region within which to calculate the filter
      coefficients.

    *noise_variance* 
      Variance of the noise in the image. If negative, estimated
      automatically as the median of local variances.
    """
    category = "Filter"
    return_type = ImageType([GREYSCALE,GREY16,FLOAT], "output")
    self_type = ImageType([GREYSCALE,GREY16,FLOAT])
    args = Args([Int("region size", default=5),
                 Real("noise variance", default=-1.0)])
    doc_examples = [(GREYSCALE,), (GREY16,), (FLOAT,)]
    def __call__(self, region_size=5, noise_variance=-1):
        return _binarization.wiener_filter(self, region_size, noise_variance)
    __call__ = staticmethod(__call__)


class niblack_threshold(PluginFunction):
    """
    Creates a binary image using Niblack's adaptive algorithm.

    Niblack, W. 1986. *An Introduction to Digital Image Processing.* Englewood
    Cliffs, NJ: Prentice Hall.

    Like the QGAR library, there are two extra global thresholds for
    the lightest and darkest regions.

    *region_size*
      The size of the region in which to calculate a threshold.

    *sensitivity*
      The sensitivity weight on the variance.

    *lower bound*
      A global threshold beneath which all pixels are considered black.

    *upper bound*
      A global threshold above which all pixels are considered white.
    """
    return_type = ImageType([ONEBIT], "output")
    self_type = ImageType([GREYSCALE])
    args = Args([Int("region size", default=15),
                 Real("sensitivity", default=-0.2),
                 Int("lower bound", range=(0,255), default=20),
                 Int("upper bound", range=(0,255), default=150)])
    doc_examples = [(GREYSCALE,)]
    def __call__(self, 
                 region_size=15, 
                 sensitivity=-0.2,
                 lower_bound=20,
                 upper_bound=150):
        return _binarization.niblack_threshold(self, 
                                               region_size, 
                                               sensitivity,
                                               lower_bound,
                                               upper_bound)
    __call__ = staticmethod(__call__)

   
class sauvola_threshold(PluginFunction):
    """
    Creates a binary image using Sauvola's adaptive algorithm.

    Sauvola, J. and M. Pietikainen. 2000. Adaptive document image
    binarization.  *Pattern Recognition* 33: 225--236.

    Like the QGAR library, there are two extra global thresholds for
    the lightest and darkest regions.

    *region_size*
      The size of the region in which to calculate a threshold.

    *sensitivity*
      The sensitivity weight on the adjusted variance.

    *dynamic_range*
      The dynamic range of the variance.

    *lower bound*
      A global threshold beneath which all pixels are considered black.

    *upper bound*
      A global threshold above which all pixels are considered white.
    """
    return_type = ImageType([ONEBIT], "output")
    self_type = ImageType([GREYSCALE])
    args = Args([Int("region size", default=15),
                 Real("sensitivity", default=0.5),
                 Int("dynamic range", range=(1, 255), default=128),
                 Int("lower bound", range=(0,255), default=20),
                 Int("upper bound", range=(0,255), default=150)])
    doc_examples = [(GREYSCALE,)]
    def __call__(self, 
                 region_size=15, 
                 sensitivity=0.5, 
                 dynamic_range=128,
                 lower_bound=20,
                 upper_bound=150):
        return _binarization.sauvola_threshold(self, 
                                               region_size, 
                                               sensitivity, 
                                               dynamic_range,
                                               lower_bound,
                                               upper_bound)
    __call__ = staticmethod(__call__)

class gatos_background(PluginFunction):
    """
    Estimates the background of an image according to Gatos et al.'s
    method. See:

    Gatos, Basilios, Ioannis Pratikakis, and Stavros
    J. Perantonis. 2004. An adaptive binarization technique for low
    quality historical documents. *Lecture Notes in Computer
    Science* 3163: 102--113.

    *region_size* 
      Region size for interpolation.

    *binarization*
      A preliminary binarization of the image.
    """
    category = "Binarization/RegionInformation"
    return_type = ImageType([GREYSCALE], "output")
    self_type = ImageType([GREYSCALE])
    args = Args([ImageType([ONEBIT], "binarization"),
                 Int("region size", default=15)])
    def __call__(self, binarization, region_size=15):
        return _binarization.gatos_background(self, binarization, region_size)
    __call__ = staticmethod(__call__)
        

class gatos_threshold(PluginFunction):
    """
    Thresholds an image according to Gatos et al.'s method. See:

    Gatos, Basilios, Ioannis Pratikakis, and Stavros
    J. Perantonis. 2004. An adaptive binarization technique for low
    quality historical documents. *Lecture Notes in Computer
    Science* 3163: 102-113.

    *background*
      Estimated background of the image.

    *binarization*
      A preliminary binarization of the image.

    Use the default settings for the other parameters unless you know
    what you are doing.
    """
    return_type = ImageType([ONEBIT], "output")
    self_type = ImageType([GREYSCALE])
    args = Args([ImageType([GREYSCALE], "background"),
                 ImageType([ONEBIT], "binarization"),
                 Real("q", default=0.6),
                 Real("p1", default=0.5),
                 Real("p2", default=0.8)])
    def __call__(self, background, binarization, q=0.6, p1=0.5, p2=0.8):
        return _binarization.gatos_threshold(self, 
                                             background, 
                                             binarization, 
                                             q, 
                                             p1, 
                                             p2)
    __call__ = staticmethod(__call__)


class white_rohrer_threshold(PluginFunction):
    """
    Creates a binary image using White and Rohrer's dynamic thresholding
    algorithm. It is the first of the two algorithms described in:

    J. M. White and G. D. Rohrer. 1983. Image thresholding for optical
    character recognition and other applications requiring character
    image extraction.  *IBM J. Res. Dev.* 27(4), pp. 400-411

    The algorithm uses a 'running' average instead of true average of
    the gray values in the neighborhood.  The lookahead parameter
    gives the number of lookahead pixels used in the biased running
    average that is used in deciding the threshold at each pixel
    location.

    *x_lookahead*
      the number of lookahead pixels in the horizontal direction for
      computing the running average. White and Rohrer suggest a value
      of 8 for a 240 dpi scanning resolution.

    *y_lookahead*
      number of lines used for further averaging from the horizontal
      averages.

    The other parameters are for calculating biased running average.
    Without bias the thresholding decision would be determined by
    noise fluctuations in uniform areas.

    This implementation uses code from XITE__.

    .. __: http://www.ifi.uio.no/forskning/grupper/dsb/Software/Xite/

    .. note::

       Permission to use, copy, modify and distribute this software
       and its documentation for any purpose and without fee is hereby
       granted, provided that this copyright notice appear in all
       copies and that both that copyright notice and this permission
       notice appear in supporting documentation and that the name of
       B-lab, Department of Informatics or University of Oslo not be
       used in advertising or publicity pertaining to distribution of
       the software without specific, written prior permission.

       B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
       INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
       FITNESS, IN NO EVENT SHALL B-LAB BE LIABLE FOR ANY SPECIAL,
       INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
       RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
       ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
       ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
       THIS SOFTWARE.
    """
    return_type = ImageType([ONEBIT], "onebit")
    self_type = ImageType([GREYSCALE])
    args = Args([Int("x lookahead", default=8),
                 Int("y lookahead", default=1),
                 Int("bias mode", default=0),
                 Int("bias factor", default=100),
                 Int("f factor",default=100),
                 Int("g factor",default=100)])
    author = "Uma Kompella (using code from the XITE library)"
    doc_examples = [(GREYSCALE,)]

    def __call__(self, x_lookahead=8, y_lookahead=1, bias_mode=0,
                 bias_factor=100, f_factor=100, g_factor=100):
        return _binarization.white_rohrer_threshold(
            self, 
            x_lookahead, 
            y_lookahead,
            bias_mode,
            bias_factor,
            f_factor,
            g_factor)
    __call__ = staticmethod(__call__)


class shading_subtraction(PluginFunction):
    """
    Thresholds an image after subtracting a -possibly shaded- background.

    First the background image is extracted with a maximum filter with a
    *k\*k* window, and this image is subtracted from the original image.
    On the difference image, a threshold is applied, and the inverted
    image thereof is the binarization result.

    Parameters:

    *k*
      Window size of the maximum filter. Must be odd. For decent results,
      it must be chosen so large that every window includes at least one
      background pixel.

    *threshold*
      Threshold applied to the difference image. A possibly reasonable
      value might lie around 20. When ``None``, the threshold is
      determined automatically with otsu_find_threshold_.

    .. _otsu_find_threshold: binarization.html#otsu-find-threshold

    Reference: K.D. Toennies: *Grundlagen der Bildverarbeitung.* 
    Pearson Studium, 2005, p.202
    """
    author = "Christoph Dalitz"
    return_type = ImageType([ONEBIT], "onebit")
    self_type = ImageType([GREYSCALE])
    args = Args([Int("k", default=7), Int("threshold", default=NoneDefault)])
    pure_python = True
    doc_examples = [(GREYSCALE,)]

    def __call__(self, k=7, threshold=None):
        #background = self.rank(k*k,k,border_treatment=1)
        background = self.min_max_filter(k,1)
        backfloat = background.to_float()
        imgfloat = self.to_float()
        difffloat = backfloat.subtract_images(imgfloat)
        if threshold is None:
            diffgrey = difffloat.to_greyscale()
            diffgrey.invert()
            return diffgrey.otsu_threshold()
        else:
            onebit = difffloat.threshold(threshold)
            onebit.invert()
            return onebit

    __call__ = staticmethod(__call__)

class brink_threshold(PluginFunction):
    """
    Calculates threshold for image with Brink and Pendock's minimum-cross    
    entropy method and returns corrected image. It is best used for binarising
    images with dark, near-black foreground and significant bleed-through.
    To that end, it generally predicts lower thresholds than other
    thresholding algorithms.

    Reference: A.D. Brink, N.E. Pendock: Minimum cross-entropy threshold selection.
    Pattern Recognition 29 (1), 1996. 179-188.
    """
    author = "Johanna Devaney, Brian Stern"
    self_type = ImageType([GREYSCALE])
    return_type = ImageType([ONEBIT], "onebit")
    doc_examples = [(GREYSCALE,)]
    def __call__(self):
        return _binarization.brink_threshold(self)
    __call__ = staticmethod(__call__)


class BinarizationGenerator(PluginModule):
    category = "Binarization"
    cpp_headers = ["binarization.hpp"]
    functions = [image_mean,
                 image_variance,
                 mean_filter,
                 variance_filter,
                 wiener_filter,
                 niblack_threshold, 
                 sauvola_threshold,
                 gatos_background,
                 gatos_threshold,
                 white_rohrer_threshold,
                 shading_subtraction,
                 brink_threshold]
    author = "John Ashley Burgoyne and Ichiro Fujinaga"
    url = "http://gamera.sourceforge.net/"

module = BinarizationGenerator()

